{*****************************************************************************}
{*                                                                           *}
{*  Program : Schedul                                                        *}
{*  Unit    : Alg 1                                                          *}
{*                                                                           *}
{*  This unit contains procedure implementing scheduling algorithm           *}
{*                                                                           *}
{*           P2 | size(j), p(j)=1, r(j) | Cmax                               *}
{*****************************************************************************}

Unit Alg1;

{$C MOVEABLE DEMANDLOAD DISCARDABLE}

Interface

Uses Deklar,WinDos,WinPROCS,WinTypes,Tools;

Procedure Alg1_(Var task: task_rec;Var comp: comp_rec; Var sched: schedul_rst;
                   Var workt: real;Var kp: ig);
{ Parameters given:
     task  - task set,
     comp - processor set,
     kp   - number of processors etc.,
  Parameters returned:
     sched   - schedule description,
     workt - how long schedule was computed}

Implementation

Type WordTable = Array [1..MaxN] Of Word;    {tablica numerow zadan}
     PWordTable = ^WordTable;
     ExtendedTable = Array [1..MaxN] Of Extended;
     PExtendedTable = ^ExtendedTable;


Procedure Alg1_(Var task: task_rec;Var comp: comp_rec; Var sched: schedul_rst;
                   Var workt: real;Var kp: ig);

Var
   i,p,k : Word;
   l, task1, task2 : Word;                            {Liczba roznych release times'ow}
   hr,mn,sc,sc100: Word;                {czas}
   t : Extended;                         {aktualny czas wykonywania zadan na procesorach}
   Release_Times : PExtendedTable;
   Cmax, Lmax, Lateness :Extended;
   BigSet, Set1, Set2 : PWordTable;
   RdTime_Tab : PExtendedTable;
   Set1_Number, Set2_Number : Word;
   Set1_Pointer, Set2_Pointer : Word;

{-----------------------------}

Procedure CorrectData;
Var i: Word;
Begin
  For i:=1 To task.n Do
  Begin
    task.task_v[i]^.tasktime := 1;
    task.task_v[i]^.ddate := Round(task.task_v[i]^.ddate) ;
    If task.task_v[i]^.ddate < 1 Then task.task_v[i]^.ddate :=1;
    task.task_v[i]^.rdytime := Round(task.task_v[i]^.rdytime);
    task.task_v[i]^.prcsnum := Round(task.task_v[i]^.prcsnum);
    If task.task_v[i]^.prcsnum < 1 Then task.task_v[i]^.prcsnum :=1;
    If task.task_v[i]^.prcsnum > 2 Then task.task_v[i]^.prcsnum :=2;
  End;
End;

{-----------------------------}

Function ready_exist(ReadyTime : Extended): Boolean;
{Ta funkcja sprawdza czy release time danego zadania
juz kiedys wystapil czy jeszcze nie - co bedzie wykorzystane
pry ustaleniu liczby interwalow czyli liczby roznych wartosci
release time }
var r:Word;
    exist:Boolean;
Begin
  r:=1;
  exist := False;
  While ((exist = False) And (r<=l)) Do
  Begin
    If ReadyTime = Release_Times^[r] Then exist := True;
    r := r + 1;
  End;
  ready_exist := exist;
End;

{-----------------------------}

Procedure Sort_All_Tasks_By_Release_Time(Var R_Set : PWordTable; num : Word);
{Procedura ta sortuje zadania wedlug czasu przybycia od najmniejszego do najwiekszego}
var i,j,min_index : Word;
    min_rdtime : Extended;
    min_name : Word;
Begin
  For i := 1 To num-1  Do
  Begin
    min_rdtime := RdTime_Tab^[R_Set^[i]];
    min_name := R_Set^[i];
    min_index := i;
    For j := (i+1) To num Do
    Begin
      if RdTime_Tab^[R_Set^[j]] < min_rdtime Then
      Begin
        min_rdtime := RdTime_Tab^[R_Set^[j]];
        min_name := R_Set^[j];
        min_index := j;
      End;
    End;
    R_Set^[min_index] := R_Set^[i];
    R_Set^[i] := min_name;
  End;
End; {Sort_All_Tasks_By_Release_Time}

{-----------------------------}

Procedure Sort_Release_Times(Var Rel_Tab:PExtendedTable);
{Procedura ta sortuje czasy przybycia od najwczesniejszego do najpozniejszego}
var i,j,actual_index : Word;
    min : Extended;
Begin
  For i := 1 To l-1  Do
  Begin
    min := Rel_Tab^[i];
    actual_index := i;
    For j := (i+1)  To l Do
    Begin
      If Rel_Tab^[j] < min Then
      Begin
        min := Rel_Tab^[j];
        actual_index := j;
      End;
    End;
    Rel_Tab^[actual_index] := Rel_Tab^[i];
    Rel_Tab^[i] := min;
  End;
End; {sort_release_times}

{-----------------------------}

Procedure Split_Tasks_Into_Two_Sets(Var LBigSet: PWordTable; Var LSet1 : PWordTable; Var LSet2 : PWordTable;
                                    Var LSet1_Number : Word; Var LSet2_Number:Word);
Var i : Word;
Begin
  LSet1_Number := 0;
  LSet2_Number := 0;
  Set1_Pointer := 1;
  Set2_Pointer := 1;
  For i := 1 To task.n Do
  Begin
    If task.task_v[LBigSet^[i]]^.prcsnum = 1 Then
     Begin
       LSet1_Number := LSet1_Number + 1;
       LSet1^[LSet1_Number] := LBigSet^[i];
     End
    Else
     Begin
       LSet2_Number := LSet2_Number + 1;
       LSet2^[LSet2_Number] := LBigSet^[i];
     End;
  End;
End;

{-----------------------------}

Function Ready_2_Task : Word;
Begin
  If (Set2_Pointer <= Set2_Number ) And
     (RdTime_Tab^[Set2^[Set2_Pointer]] < Release_Times^[k+1]) And
     (Set2_Pointer <= Set2_Number)
  Then Ready_2_Task := Set2^[Set2_Pointer]
  Else Ready_2_Task := 0;
End;

{-----------------------------}

Function Ready_1_Task : Word;
Begin
  If (Set1_Pointer <= Set1_Number) And
     (RdTime_Tab^[Set1^[Set1_Pointer]] < Release_Times^[k+1]) And
     (Set1_Pointer <= Set1_Number)
  Then Ready_1_Task := Set1^[Set1_Pointer]
  Else Ready_1_Task := 0;
End;

{-----------------------------}

Function Interval_Is_Empty : Boolean;
Var x:Word;
Begin
  Interval_Is_Empty := True;
  If ((Ready_2_Task > 0) Or (Ready_1_Task > 0)) Then Interval_Is_Empty := False;
End;

{-----------------------------}

Procedure DecreaseSet1;
Begin
  Set1^[Set1_Pointer] := 0;
  Set1_Pointer := Set1_Pointer + 1;
End;

{-----------------------------}

Procedure DecreaseSet2;
Begin
  Set2^[Set2_Pointer] := 0;
  Set2_Pointer := Set2_Pointer + 1;
End;

{-----------------------------}

Begin               {Alg 1}
    memory:=0;
    New(Release_Times);
    New(Set1);
    New(Set2);
    New(BigSet);
    New(RdTime_Tab);
    With comp Do
      With sched Do
      Begin
        initsched(sched,1,nr_comp_proc);
        For i:=1 To nr_comp_proc Do schedul[i]^.nrts:=0;
      End; {with}
    sched.m := comp.nr_comp_proc;
    CorrectData;
    For i:=1 To task.n Do
    Begin
      BigSet^[i] := i;
      RdTime_Tab^[i] := task.task_v[i]^.rdytime;
      Set1^[i] := 0;
      Set2^[i] := 0;
    End;
    Set1_Number := 0;
    Set2_Number := 0;
    GetTime(hr,mn,sc,sc100);
    Lmax := -1234567;
    Sort_All_Tasks_By_Release_Time(BigSet, task.n);
    Split_Tasks_Into_Two_Sets(BigSet, Set1, Set2, Set1_Number, Set2_Number);
    l:=0;
    For p:=1 To task.n Do Release_Times^[p]:= -1;
    For p:=1 To task.n Do {We'll find out how many different values of release times is present}
    Begin
      IF ready_exist(task.task_v[p]^.rdytime) = False Then
      Begin
        l := l+1;
        release_times^[l] := task.task_v[p]^.rdytime ;
      End;
    End; {l - number of different values of release times}
         {so l is number of intervals}
         {release_times^ - table with these values}
    Sort_Release_Times(Release_Times);
    Release_Times^[l+1] := 21474;
    t := 0;
    t := Release_Times^[1];
    memory:=memory+SizeOf(Release_Times^)+SizeOf(Set1^)+SizeOf(Set2^)+
                  +SizeOf(BigSet^)+SizeOf(RdTime_Tab^);

    For k:=1 To l Do
    Begin
      While (Release_Times^[k+1] > t) And (Interval_Is_Empty = False) Do
      Begin
        Task2 := Ready_2_Task;
        Task1 := Ready_1_Task;
        If Task2 > 0 Then
        Begin
          i := sched.schedul[1]^.nrts + 1;
          sched.schedul[1]^.a[i].tnr:=Task2;
          sched.schedul[1]^.a[i].begint:=t;
          sched.schedul[1]^.a[i].endt:=t+1;
          sched.schedul[1]^.a[i].usestime:=1;
          sched.schedul[1]^.end_time:=t+1;
          sched.schedul[1]^.nrts := sched.schedul[1]^.nrts+1;
          i := sched.schedul[2]^.nrts + 1;
          sched.schedul[2]^.a[i].tnr:=Task2;
          sched.schedul[2]^.a[i].begint:=t;
          sched.schedul[2]^.a[i].endt:=t+1;
          sched.schedul[2]^.a[i].usestime:=1;
          sched.schedul[2]^.end_time:=t+1;
          sched.schedul[2]^.nrts:=sched.schedul[2]^.nrts+1;
          Lateness := (t + 1) - task.task_v[Task2]^.ddate;
          If Lateness > Lmax Then Lmax := Lateness;
          DecreaseSet2;
        End
        Else If Task1 > 0 Then
        Begin
          i := sched.schedul[1]^.nrts + 1;
          sched.schedul[1]^.a[i].tnr:=Task1;
          sched.schedul[1]^.a[i].begint:=t;
          sched.schedul[1]^.a[i].endt:=t+1;
          sched.schedul[1]^.a[i].usestime:=1;
          sched.schedul[1]^.end_time:=t+1;
          sched.schedul[1]^.nrts:=sched.schedul[1]^.nrts+1;
          Lateness := (t + 1) - task.task_v[Task1]^.ddate;
          If Lateness > Lmax Then Lmax := Lateness;
          DecreaseSet1;
          Task1 := Ready_1_Task;
          If Task1 > 0 Then
          Begin
            i := sched.schedul[2]^.nrts + 1;
            sched.schedul[2]^.a[i].tnr:=Task1;
            sched.schedul[2]^.a[i].begint:=t;
            sched.schedul[2]^.a[i].endt:=t+1;
            sched.schedul[2]^.a[i].usestime:=1;
            sched.schedul[2]^.end_time:=t+1;
            sched.schedul[2]^.nrts:=sched.schedul[2]^.nrts+1;
            Lateness := (t + 1) - task.task_v[Task1]^.ddate;
            If Lateness > Lmax Then Lmax := Lateness;
            DecreaseSet1;
          End;
        End;
        t := t + 1;
      End; {While loop}
      If Interval_Is_Empty = True Then
      Begin
        If (k < l)
        Then If (Release_Times^[k+1] > t)
             Then t := Release_Times^[k+1];
      End;
    End; {For loop}
    sched.c:=t;
    sched.Lmax := Lmax;
    sched.m:=comp.nr_comp_proc;
    workt:=GetSec(hr,mn,sc,sc100);
    Dispose(Set1);
    Dispose(Set2);
    Dispose(RdTime_Tab);
    Dispose(BigSet);
    Dispose(Release_Times);
    is_act_sch:=true;
  End;
End.  {Alg 1}