{*****************************************************************************}
{*                                                                           *}
{*  Program : Schedul                                                        *}
{*  Unit    : Alg3 (Based On Hu Algorithm)                                   *}
{*                                                                           *}
{*  This unit contains procedure implementing                                *}
{*  P2|size(j), p(j)=1, tree|Cmax  scheduling algorithm.                     *}
{*                                                                           *}
{*****************************************************************************}

Unit Alg3;

{$C MOVEABLE DEMANDLOAD DISCARDABLE}

Interface

Uses Deklar,WinDos,WinPROCS,WinTypes,Tools;

Procedure Alg3_(Var task: task_rec;Var comp: comp_rec; Var sched: schedul_rst;
                   Var workt: real;Var kp: ig);
{ Parameters given:
     task  - task set,
     comp - processor set,
     kp   - number of processors etc.,
  Parameters returned:
     sched   - schedule description,
     workt - how long schedule was computed}

Implementation

Type Tlevel = Array [0..MaxN] Of Integer;    {tablica poziomow zadan}
     Plevel = ^Tlevel;
     Tsucc = Array [1..MaxN] Of  fs;      {tablica zbiorow nastepnikow zadan}
     Psucc = ^Tsucc;
     Tno_exec = Array [1..MaxN] Of fs;    {tablica zbiorow nie wykonanych poprzednikow}
     Pno_exec = ^Tno_exec;

Procedure Alg3_(Var task: task_rec;Var comp: comp_rec; Var sched: schedul_rst;
                   Var workt: real;Var kp: ig);

Var
   i,j,k: Byte;                       {identyfikatory elementow tablic}
   level: Plevel;
   MaxLate, Lateness : Real;
   hr,mn,sc,sc100: Word;              {czas}
   non_assigned: fs;                  {zb. zad. nieprzydzielonych}
   execut: ^fs;
   succ: Psucc;
   found : Boolean;
   no_exec: Pno_exec;
   wsk: ptrprem;
   t: Byte;                           {aktualny czas wykonywania zadan na obu maszynach}
   h: Byte;                           {ident. przydzielanych zadan}
   TaskName, FirstTask, SecondTask, index : Word;

{-----------------------------------------------------}

Procedure set_levels(n: Byte; Var level: Plevel; Var succ: Psucc);
{This procedure sets levels of tasks}
Var
   lev,i,j: Byte;
   not_desc: fs;    {zb.zad., ktorym nie okreslono poziomu}
   desc: fs;        {zb.zad., ktorym okreslono poziom w danej iteracji}
Begin
  desc:=[];
  not_desc:=non_assigned;
  For i:=0 To n Do level^[i]:=0;
  lev:=1;
  While not_desc<>[] Do
  Begin
    found := true;
    While found = true Do
    Begin {Remove recursively ready 2-tasks}
      found := false;
      For i:=1 To n Do
      Begin
        If ((level^[i]=0) And (succ^[i]=[]) And (task.task_v[i]^.prcsnum = 2)) Then
        Begin
          level^[i]:= -1;
          not_desc:=not_desc-[i];
          desc:=desc+[i];
          found := true;
        End;  {If}
      End; {For}
      For i:=1 To n Do
      If (i in desc) Then Begin
                            For j:=1 To n Do succ^[j]:=succ^[j]-[i];
                            desc:=desc-[i];
                          End;
    End; {Remove recursively ready 2-tasks}

    Begin {Assign levels to ready 1-tasks}
      For i:=1 To n Do
      Begin
        If ((level^[i]=0) And (succ^[i]=[])) Then
        Begin
          level^[i]:=lev;
          not_desc:=not_desc-[i];
          desc:=desc+[i];
        End;  {If}
      End; {For}
      For i:=1 To n Do
      If (i in desc) Then Begin
                            For j:=1 To n Do succ^[j]:=succ^[j]-[i];
                            desc:=desc-[i];
                          End;
      lev:=lev+1;
    End; {Assign levels to ready 1-tasks}
  End; {While}
End; {set_levels}

{-----------------------------------------------------}

Function Find_Ready_2_Task : Word;
Var Name, q : Word;
Begin
  Name := 0;
  q := 1;
  While ((Name = 0) And (q <= task.n)) Do
  Begin
    If ((q in non_assigned) And (no_exec^[q]=[])
    And (Not(q in execut^)) And (task.task_v[q]^.prcsnum = 2)) Then Name := q;
    q := q + 1;
  End;
  Find_Ready_2_Task := Name;
End;

{-----------------------------------------------------}

Procedure Find_Highest_1_Tasks(Var FirstTask : Word; Var SecondTask : Word);
Var Highest, q : Word;
Begin
  Highest := 0;
  For q:=1 To task.n Do
    If ((q in non_assigned) And (level^[q]>level^[Highest]) And (no_exec^[q]=[])
    And (Not(q in execut^)) And (task.task_v[q]^.prcsnum = 1))Then Highest:=q;
  FirstTask := Highest;
  Highest := 0;
  For q:=1 To task.n Do
    If ((q <> FirstTask) And (q in non_assigned) And (level^[q]>level^[Highest]) And (no_exec^[q]=[])
    And (Not(q in execut^)) And (task.task_v[q]^.prcsnum = 1))Then Highest:=q;
  SecondTask := Highest;
End;

{-----------------------------}

Begin               {Alg3}
    memory:=0;
    MaxLate := -123456;
    New(level);
    New(succ);
    New(no_exec);
    New(execut);
    non_assigned:=[];                      {Tworzenie zbioru non_assigned i tablicy no_exec}
    For i:=1 To task.n Do
    Begin
      no_exec^[i]:=[];
      succ^[i]:=[];
    End;
    For i:=1 To task.n Do
    Begin
      non_assigned:=non_assigned+[i];
      If task.task_v[i]^.next<>nil Then
      Begin
        wsk:=task.task_v[i]^.next;
        While wsk<>nil Do
        Begin
          no_exec^[wsk^.nrprem]:=no_exec^[wsk^.nrprem]+[i]; {zad i jako poprzednik wsk^.nrprm}
          succ^[i]:=succ^[i]+[wsk^.nrprem];                 {wpis do tablicy nastepnikow}
          wsk:=wsk^.next;                                   {nastepny element a listy}
        End; {While}
      End; {If}
    End; {For}

    For i:=1 To task.n Do memory:=memory+SizeOf(no_exec^[i])+SizeOf(succ^[i]);
    memory:=memory+SizeOf(non_assigned)+SizeOf(execut^)+SizeOf(level^);

    For i:=1 To task.n Do
    Begin
      If task.task_v[i]^.prcsnum < 1 Then task.task_v[i]^.prcsnum := 1;
      If task.task_v[i]^.prcsnum > 2 Then task.task_v[i]^.prcsnum := 2;
    End;

    t:=0;     {Ustawienie poczatkowych wartosci}
    With comp Do
      With sched Do
      Begin
        initsched(sched,1,nr_comp_proc);
        For i:=1 To nr_comp_proc Do schedul[i]^.nrts:=0;
      End; {with}
    i:=1;

    GetTime(hr,mn,sc,sc100);
    set_levels(task.n, level, succ);   {Okreslanie poziomow zadan}

    While non_assigned<>[] Do          {Przydzielenie zadan do procesorow}
    Begin
      execut^:=[];
      {Schedule All Ready 2-tasks}
      Taskname := Find_Ready_2_Task;
      While Taskname <> 0 Do
      Begin
        sched.schedul[1]^.nrts := sched.schedul[1]^.nrts + 1;
        index := sched.schedul[1]^.nrts;
        sched.schedul[1]^.a[index].tnr := Taskname;
        sched.schedul[1]^.a[index].begint:=t;
        sched.schedul[1]^.a[index].endt:=t+1;
        sched.schedul[1]^.a[index].usestime:=1;
        sched.schedul[1]^.end_time:=t+1;
        sched.schedul[2]^.nrts := sched.schedul[2]^.nrts + 1;
        index := sched.schedul[2]^.nrts;
        sched.schedul[2]^.a[index].tnr := Taskname;
        sched.schedul[2]^.a[index].begint:=t;
        sched.schedul[2]^.a[index].endt:=t+1;
        sched.schedul[2]^.a[index].usestime:=1;
        sched.schedul[2]^.end_time:=t+1;
        non_assigned:=non_assigned-[Taskname];
        execut^:=execut^+[Taskname];
        t := t + 1;
        Lateness := t - task.task_v[Taskname]^.ddate;
        If MaxLate < Lateness Then MaxLate := Lateness;
        For k:=1 To task.n Do no_exec^[k] := no_exec^[k]-[Taskname];
        Taskname := Find_Ready_2_Task;
      End;{End Of Schedule All Ready 2-tasks}

      {Schedule 2 highest 1-tasks if exist}
      Find_Highest_1_Tasks(FirstTask, SecondTask);
      If FirstTask <> 0 Then
      Begin
        sched.schedul[1]^.nrts := sched.schedul[1]^.nrts + 1;
        index := sched.schedul[1]^.nrts;
        sched.schedul[1]^.a[index].tnr := FirstTask;
        sched.schedul[1]^.a[index].begint:=t;
        sched.schedul[1]^.a[index].endt:=t+1;
        sched.schedul[1]^.a[index].usestime:=1;
        sched.schedul[1]^.end_time:=t+1;
        non_assigned:=non_assigned-[FirstTask];
        execut^:=execut^+[FirstTask];
        Lateness := (t+1) - task.task_v[FirstTask]^.ddate;
        If MaxLate < Lateness Then MaxLate := Lateness;
        If SecondTask <> 0 Then
        Begin
          sched.schedul[2]^.nrts := sched.schedul[2]^.nrts + 1;
          index := sched.schedul[2]^.nrts;
          sched.schedul[2]^.a[index].tnr := SecondTask;
          sched.schedul[2]^.a[index].begint:=t;
          sched.schedul[2]^.a[index].endt:=t+1;
          sched.schedul[2]^.a[index].usestime:=1;
          sched.schedul[2]^.end_time:=t+1;
          non_assigned:=non_assigned-[SecondTask];
          execut^:=execut^+[SecondTask];
          Lateness := (t+1) - task.task_v[SecondTask]^.ddate;
          If MaxLate < Lateness Then MaxLate := Lateness;
        End;
      t := t + 1;
      End;

      {Remove scheduled tasks from "no_exec"}
      For j:=1 To task.n Do
        If (j in execut^) Then For k:=1 To task.n Do no_exec^[k]:=no_exec^[k]-[j];
    End; {While}

    sched.c:=t;
    sched.Lmax := MaxLate;
    sched.m:=comp.nr_comp_proc;
    workt:=GetSec(hr,mn,sc,sc100);
    Dispose(execut);
    Dispose(no_exec);
    Dispose(succ);
    Dispose(level);
    is_act_sch:=true;
  End;  {Alg3}
End.



