{*****************************************************************************}
{*                                                                           *}
{*  Program : Schedul                                                        *}
{*  Unit    : Alg4                                                           *}
{*                                                                           *}
{*  This unit contains procedure implementing                                *}
{*  P2|size(j), p(j)=1, r(j), tree|Cmax  scheduling algorithm.               *}
{*                                                                           *}
{*****************************************************************************}

Unit Alg4;

{$C MOVEABLE DEMANDLOAD DISCARDABLE}

Interface

Uses Deklar,WinDos,WinPROCS,WinTypes,Tools;

Procedure Alg4_(Var task: task_rec;Var comp: comp_rec; Var sched: schedul_rst;
                   Var workt: real;Var kp: ig);
{ Parameters given:
     task  - task set,
     comp - processor set,
     kp   - number of processors etc.,
  Parameters returned:
     sched   - schedule description,
     workt - how long schedule was computed}

Implementation

Type Tlevel = Array [0..MaxN] Of Integer;    {tablica poziomow zadan}
     Plevel = ^Tlevel;
     Tsucc = Array [1..MaxN] Of  fs;      {tablica zbiorow nastepnikow zadan}
     Psucc = ^Tsucc;
     Tno_exec = Array [1..MaxN] Of fs;    {tablica zbiorow nie wykonanych poprzednikow}
     Pno_exec = ^Tno_exec;
     RealTable = Array [1..MaxN] Of Real;
     PRealTable = ^RealTable;


Procedure Alg4_(Var task: task_rec;Var comp: comp_rec; Var sched: schedul_rst;
                   Var workt: real;Var kp: ig);

Var
   i,j,k: Byte;                       {identyfikatory elementow tablic}
   l, p, interval : Word;
   level: Plevel;
   MaxLate, Lateness : Real;
   hr,mn,sc,sc100: Word;              {czas}
   non_assigned: fs;                  {zb. zad. nieprzydzielonych}
   execut: ^fs;
   succ: Psucc;
   found : Boolean;
   no_exec: Pno_exec;
   wsk: ptrprem;
   t: Real;                           {aktualny czas wykonywania zadan na obu maszynach}
   h: Byte;                           {ident. przydzielanych zadan}
   TaskName, FirstTask, SecondTask, index : Word;
   Release_Times : PRealTable;
   RdTime_Tab : PRealTable;
   Interval_Is_Empty : Boolean;

{-----------------------------------------------------}

Procedure CorrectData;
Var i: Word;
Begin
  For i:=1 To task.n Do
  Begin
    task.task_v[i]^.tasktime := 1;
    task.task_v[i]^.ddate := Round(task.task_v[i]^.ddate) ;
    If task.task_v[i]^.ddate < 0 Then task.task_v[i]^.ddate :=0;
    task.task_v[i]^.rdytime := Round(task.task_v[i]^.rdytime);
    task.task_v[i]^.prcsnum := Round(task.task_v[i]^.prcsnum);
    If task.task_v[i]^.prcsnum < 1 Then task.task_v[i]^.prcsnum :=1;
    If task.task_v[i]^.prcsnum > 2 Then task.task_v[i]^.prcsnum :=2;
  End;
End;

{----------------------------------------------------}

Function ready_exist(ReadyTime : Extended): Boolean;
{Ta funkcja sprawdza czy release time danego zadania
juz kiedys wystapil czy jeszcze nie - co bedzie wykorzystane
pry ustaleniu liczby interwalow czyli liczby roznych wartosci
release time }
var r:Word;
    exist:Boolean;
Begin
  r:=1;
  exist := False;
  While ((exist = False) And (r<=l)) Do
  Begin
    If ReadyTime = Release_Times^[r] Then exist := True;
    r := r + 1;
  End;
  ready_exist := exist;
End;

{-----------------------------------------------------}

Procedure Sort_Release_Times(Var Rel_Tab:PRealTable);
{Procedura ta sortuje czasy przybycia od najwczesniejszego do najpozniejszego}
var i,j,actual_index : Word;
    min : Real;
Begin
  For i := 1 To l-1  Do
  Begin
    min := Rel_Tab^[i];
    actual_index := i;
    For j := (i+1)  To l Do
    Begin
      If Rel_Tab^[j] < min Then
      Begin
        min := Rel_Tab^[j];
        actual_index := j;
      End;
    End;
    Rel_Tab^[actual_index] := Rel_Tab^[i];
    Rel_Tab^[i] := min;
  End;
End; {sort_release_times}

{-----------------------------------------------------}

Procedure set_levels(n: Byte; Var level: Plevel; Var succ: Psucc);
{This procedure sets levels of tasks}
Var
   lev,i,j: Byte;
   not_desc: fs;    {zb.zad., ktorym nie okreslono poziomu}
   desc: fs;        {zb.zad., ktorym okreslono poziom w danej iteracji}
Begin
  desc:=[];
  not_desc:=non_assigned;
  For i:=0 To n Do level^[i]:=0;
  lev:=1;
  While not_desc<>[] Do
  Begin
    found := true;
    While found = true Do
    Begin {Remove recursively ready 2-tasks}
      found := false;
      For i:=1 To n Do
      Begin
        If ((level^[i]=0) And (succ^[i]=[]) And (task.task_v[i]^.prcsnum = 2)) Then
        Begin
          level^[i]:= -1;
          not_desc:=not_desc-[i];
          desc:=desc+[i];
          found := true;
        End;  {If}
      End; {For}
      For i:=1 To n Do
      If (i in desc) Then Begin
                            For j:=1 To n Do succ^[j]:=succ^[j]-[i];
                            desc:=desc-[i];
                          End;
    End; {Remove recursively ready 2-tasks}

    Begin {Assign levels to ready 1-tasks}
      For i:=1 To n Do
      Begin
        If ((level^[i]=0) And (succ^[i]=[])) Then
        Begin
          level^[i]:=lev;
          not_desc:=not_desc-[i];
          desc:=desc+[i];
        End;  {If}
      End; {For}
      For i:=1 To n Do
      If (i in desc) Then Begin
                            For j:=1 To n Do succ^[j]:=succ^[j]-[i];
                            desc:=desc-[i];
                          End;
      lev:=lev+1;
    End; {Assign levels to ready 1-tasks}
  End; {While}
End; {set_levels}

{-----------------------------------------------------}

Function Find_Ready_2_Task : Word;
Var Name, q : Word;
Begin
  Name := 0;
  q := 1;
  While ((Name = 0) And (q <= task.n)) Do
  Begin
    If ((q in non_assigned) And (no_exec^[q]=[])
       And (Not(q in execut^)) And (task.task_v[q]^.prcsnum = 2)
       And (task.task_v[q]^.rdytime <= Release_Times^[interval]))
    Then Name := q;
    q := q + 1;
  End;
  Find_Ready_2_Task := Name;
End;

{-----------------------------------------------------}

Procedure Find_Highest_1_Tasks(Var FirstTask : Word; Var SecondTask : Word);
Var Highest, q : Word;
Begin
  Highest := 0;
  For q:=1 To task.n Do
    If ((q in non_assigned) And (level^[q]>level^[Highest]) And (no_exec^[q]=[])
       And (Not(q in execut^)) And (task.task_v[q]^.prcsnum = 1)
       And (task.task_v[q]^.rdytime <= Release_Times^[interval]))
    Then Highest:=q;
  FirstTask := Highest;
  Highest := 0;
  For q:=1 To task.n Do
    If ((q <> FirstTask) And (q in non_assigned) And (level^[q]>level^[Highest]) And (no_exec^[q]=[])
       And (Not(q in execut^)) And (task.task_v[q]^.prcsnum = 1)
       And (task.task_v[q]^.rdytime <= Release_Times^[interval]))
    Then Highest:=q;
  SecondTask := Highest;
End;

{-----------------------------}

Begin               {Alg4}
    memory:=0;
    MaxLate := -123456;
    New(level);
    New(succ);
    New(no_exec);
    New(execut);
    New(Release_Times);
    New(RdTime_Tab);
    non_assigned:=[];                      {Tworzenie zbioru non_assigned i tablicy no_exec}
    For i:=1 To task.n Do
    Begin
      no_exec^[i]:=[];
      succ^[i]:=[];
    End;
    For i:=1 To task.n Do
    Begin
      non_assigned:=non_assigned+[i];
      If task.task_v[i]^.next<>nil Then
      Begin
        wsk:=task.task_v[i]^.next;
        While wsk<>nil Do
        Begin
          no_exec^[wsk^.nrprem]:=no_exec^[wsk^.nrprem]+[i]; {zad i jako poprzednik wsk^.nrprm}
          succ^[i]:=succ^[i]+[wsk^.nrprem];                 {wpis do tablicy nastepnikow}
          wsk:=wsk^.next;                                   {nastepny element a listy}
        End; {While}
      End; {If}
    End; {For}

    For i:=1 To task.n Do memory:=memory+SizeOf(no_exec^[i])+SizeOf(succ^[i]);
    memory:=memory+SizeOf(non_assigned)+SizeOf(execut^)+SizeOf(level^)+
            +SizeOf(Release_Times^)+SizeOf(RdTime_Tab^);

    With comp Do
      With sched Do
      Begin
        initsched(sched,1,nr_comp_proc);
        For i:=1 To nr_comp_proc Do schedul[i]^.nrts:=0;
      End; {with}

    CorrectData;
    For i:=1 To task.n Do RdTime_Tab^[i] := task.task_v[i]^.rdytime;
    GetTime(hr,mn,sc,sc100);

    l:=0;
    For p:=1 To task.n Do Release_Times^[p]:= -1;
    For p:=1 To task.n Do {We'll find out how many different values of release times is present}
    Begin
      IF ready_exist(task.task_v[p]^.rdytime) = False Then
      Begin
        l := l+1;
        release_times^[l] := task.task_v[p]^.rdytime ;
      End;
    End; {l - number of different values of release times}
         {so l is number of intervals}
         {release_times^ - table with these values}
    Sort_Release_Times(Release_Times);
    Release_Times^[l+1] := 21474;
    t := 0;
    interval := 1;
    t := Release_Times^[interval]; {Ustawienie poczatkowych wartosci}

    set_levels(task.n, level, succ);   {Okreslanie poziomow zadan}

    While non_assigned<>[] Do          {Przydzielenie zadan do procesorow}
    Begin
      execut^:=[];
      {Schedule All Ready 2-tasks}
      Taskname := Find_Ready_2_Task;
      If Taskname = 0 Then Interval_Is_Empty := True;
      While Taskname <> 0 Do
      Begin
        sched.schedul[1]^.nrts := sched.schedul[1]^.nrts + 1;
        index := sched.schedul[1]^.nrts;
        sched.schedul[1]^.a[index].tnr := Taskname;
        sched.schedul[1]^.a[index].begint:=t;
        sched.schedul[1]^.a[index].endt:=t+1;
        sched.schedul[1]^.a[index].usestime:=1;
        sched.schedul[1]^.end_time:=t+1;
        sched.schedul[2]^.nrts := sched.schedul[2]^.nrts + 1;
        index := sched.schedul[2]^.nrts;
        sched.schedul[2]^.a[index].tnr := Taskname;
        sched.schedul[2]^.a[index].begint:=t;
        sched.schedul[2]^.a[index].endt:=t+1;
        sched.schedul[2]^.a[index].usestime:=1;
        sched.schedul[2]^.end_time:=t+1;
        non_assigned:=non_assigned-[Taskname];
        execut^:=execut^+[Taskname];
        t := t + 1;
        Lateness := t - task.task_v[Taskname]^.ddate;
        If MaxLate < Lateness Then MaxLate := Lateness;
        For k:=1 To task.n Do no_exec^[k] := no_exec^[k]-[Taskname];
        Taskname := Find_Ready_2_Task;
      End;{End Of Schedule All Ready 2-tasks}

      {Schedule 2 highest 1-tasks if exist}
      Find_Highest_1_Tasks(FirstTask, SecondTask);
      If FirstTask <> 0 Then
      Begin
        Interval_Is_Empty := False;
        sched.schedul[1]^.nrts := sched.schedul[1]^.nrts + 1;
        index := sched.schedul[1]^.nrts;
        sched.schedul[1]^.a[index].tnr := FirstTask;
        sched.schedul[1]^.a[index].begint:=t;
        sched.schedul[1]^.a[index].endt:=t+1;
        sched.schedul[1]^.a[index].usestime:=1;
        sched.schedul[1]^.end_time:=t+1;
        non_assigned:=non_assigned-[FirstTask];
        execut^:=execut^+[FirstTask];
        Lateness := (t+1) - task.task_v[FirstTask]^.ddate;
        If MaxLate < Lateness Then MaxLate := Lateness;
        If SecondTask <> 0 Then
        Begin
          sched.schedul[2]^.nrts := sched.schedul[2]^.nrts + 1;
          index := sched.schedul[2]^.nrts;
          sched.schedul[2]^.a[index].tnr := SecondTask;
          sched.schedul[2]^.a[index].begint:=t;
          sched.schedul[2]^.a[index].endt:=t+1;
          sched.schedul[2]^.a[index].usestime:=1;
          sched.schedul[2]^.end_time:=t+1;
          non_assigned:=non_assigned-[SecondTask];
          execut^:=execut^+[SecondTask];
          Lateness := (t+1) - task.task_v[SecondTask]^.ddate;
          If MaxLate < Lateness Then MaxLate := Lateness;
        End;
      t := t + 1;
      End;

      {Remove scheduled tasks from "no_exec"}
      For j:=1 To task.n Do
        If (j in execut^) Then For k:=1 To task.n Do no_exec^[k]:=no_exec^[k]-[j];

      If (Interval_Is_Empty = True) Or (t >= Release_times^[interval+1]) Then
      Begin
        If (interval < l)
        Then
        Begin
          interval := interval + 1;
          If (t < Release_Times^[interval])
          Then t := Release_Times^[interval];
        End;
      End;

    End; {While}

    sched.c:=t;
    sched.Lmax := MaxLate;
    sched.m:=comp.nr_comp_proc;
    workt:=GetSec(hr,mn,sc,sc100);
    Dispose(execut);
    Dispose(no_exec);
    Dispose(succ);
    Dispose(level);
    Dispose(Release_Times);
    Dispose(RdTime_Tab);
    is_act_sch:=true;
  End;  {Alg4}
End.



