{*****************************************************************************}
{*                                                                           *}
{*  Program : Simul                                                          *}
{*  Unit    : Alg 5.20 (Badania operacyjne dla informatykow -                *}
{*                     J.Blazewicz, W.Cellary, R.Slowinski, J.Weglarz)       *}
{*  This unit contains procedure implementing 5.20 scheduling algorithm.     *}
{*                                                                           *}
{*****************************************************************************}

Unit Alg520;

{$C MOVEABLE DEMANDLOAD DISCARDABLE}

Interface

Uses Deklar,WinDos,WinPROCS,WinTypes,Tools;

Procedure Alg520_(Var task: task_rec;Var comp: comp_rec; Var sched: schedul_rst;
                   Var workt: real;Var kp: ig);
{ Parameters given:
     task  - task set,
     comp - processor set,
     kp   - number of processors etc.,
  Parameters returned:
     sched   - schedule description,
     workt - how long schedule was computed}

Implementation

Type Tsucc = Array [1..MaxN] Of  fs;      {tablica zbiorow nastepnikow zadan}
     Psucc = ^Tsucc;
     Tno_exec = Array [1..MaxN] Of fs;    {tablica zbiorow nie wykonanych poprzednikow}
     Pno_exec = ^Tno_exec;
     Ttztzz = Array [0..MaxN] Of Real;       {tablica zmodyfikowanych terminow zakonczenia}
     Ptztzz = ^Ttztzz;
     Tsorted = Array [1..MaxN] Of Real;   {tablica posortowanych zadan wg. ztz }
     Psorted = ^Tsorted;
     Ttemporal = Array [1..MaxN] Of Byte;
     Ptemporal = ^Ttemporal;

Procedure Alg520_(Var task: task_rec;Var comp: comp_rec; Var sched: schedul_rst;
                   Var workt: real;Var kp: ig);

Var
   i,j,k: Byte;                       {identyfikatory elementow tablic}
   hr,mn,sc,sc100: Word;              {czas}
   non_assigned: fs;              {zb. zad. nieprzydzielonych}
   execut: ^fs;
   succ: Psucc;
   no_exec: Pno_exec;
   wsk: ptrprem;
   t: Byte;                           {aktualny czas wykonywania zadan na obu maszynach}
   h: Byte;                           {ident. przydzielanych zadan}
   tztzz: Ptztzz;
   sorted_tztzz: Psorted;
   indexed_tab : Ptemporal;
   Lateness, MaxLate : Real;

{-----------------------------}

Function max(i: Real; j: Real) : Real;
Begin
  If (i > j) Then max := i
  Else max :=j;
End; {max}

{-----------------------------}

Procedure fill_tztzz(Var tztzz_tab: Ptztzz);
{Procedura ta oblicza zmodyfikowane terminy }
{zakonczenia dla kazdego zadania i zwraca je}
{w tablicy tztzz                            }
var i : Byte;
    a,b : Real;
Begin
  tztzz_tab^[0] := 0;
  tztzz_tab^[1] := 1 - task.task_v[1]^.ddate;
  For i := 2 To task.n Do
  Begin
    If task.task_v[i]^.next <> NIL Then a := tztzz_tab^[task.task_v[i]^.next^.nrprem]
    Else a:= 0;
    b := task.task_v[i]^.ddate;
    tztzz_tab^[i] := max(1 + a , 1 - b);
  End;
End; {fill_tztzz}

{-----------------------------}

Procedure sort_tztzz(Var tztzz_tab: Ptztzz; Var sort_tab: Psorted; Var index_tab: Ptemporal); {Posortowanie zadan wg ztz}
{Procedura ta sortuje zadania wedlug  }
{zmodyfikowanego terminu zakonczenia i}
{umieszcza numery posortowanych zadan }
{w tablicy sorted_tztzz               }
var i,j,actual_index : Byte;
    maks : record
      ddate : Real;
      index : Byte;
    end;
Begin
  For i := 1 To task.n Do
    sort_tab^[i] := tztzz_tab^[i];
  For i := 1 To task.n Do
    index_tab^[i] := i;

  For i := 1 To (task.n - 1)  Do
  Begin
    maks.ddate := sort_tab^[i];
    maks.index := index_tab^[i];
    actual_index := i;
    For j := (i+1)  To task.n Do
    Begin
      if (sort_tab^[j] > maks.ddate) Then
      Begin
        maks.ddate := sort_tab^[j];
        maks.index := index_tab^[j];
        actual_index := j;
      End;
    End;
    sort_tab^[actual_index] := sort_tab^[i];
    index_tab^[actual_index] := index_tab^[i];
    sort_tab^[i] := maks.ddate;
    index_tab^[i] := maks.index;
  End;
End; {sort_tztzz}

{-----------------------------}

Procedure high_moddate(Var x: Byte; Var no_exec:Pno_exec);
{ Procedura ta zwraca numer zadania, ktore ma najwyzszy      }
{ zmodyfikowany termin zakonczenia, nalezy do zbioru         }
{ nie przydzielonych zadan, ktorego wszystkie poprzedniki    }
{ zostaly wykonane i ktore samo nie zostalo jeszcze wykonane.}
Var
   highest,q,y: Byte;
Begin
  highest:=0;
  For y:=task.n Downto 1 Do
  Begin
    q:=indexed_tab^[y];
    If  (q in non_assigned) And
        (no_exec^[q]=[])    And
        (Not(q in execut^))
    Then highest:=q;
  End;
  x:=highest;
End; {highmoddate}


{-----------------------------}

Begin               {Alg 5.20}
  If HuConstrain<>outtree Then
  Begin
    memory:=0;
    New(indexed_tab);
    New(sorted_tztzz);
    New(tztzz);
    New(succ);
    New(no_exec);
    New(execut);
    non_assigned:=[];
    MaxLate := -123456;
    For i:=1 To task.n Do
    Begin
      no_exec^[i]:=[];
      succ^[i]:=[];
    End;
    For i:=1 To task.n Do
    Begin
      non_assigned:=non_assigned+[i];
      If task.task_v[i]^.next<>nil Then
      Begin
        wsk:=task.task_v[i]^.next;
        While wsk<>nil Do
        Begin
          no_exec^[wsk^.nrprem]:=no_exec^[wsk^.nrprem]+[i]; {zad i jako poprzednik wsk^.nrprm}
          succ^[i]:=succ^[i]+[wsk^.nrprem];                 {wpis do tablicy nastepnikow}
          wsk:=wsk^.next;                                   {nastepny element a listy}
        End; {While}
      End; {If}
    End; {For}

    For i:=1 To task.n Do memory:=memory+SizeOf(no_exec^[i])+SizeOf(succ^[i]);
    memory:=memory+SizeOf(non_assigned)+SizeOf(execut^)+SizeOf(sorted_tztzz^)+SizeOf(indexed_tab^)+SizeOf(tztzz^);

    t:=0;     {Ustawienie poczatkowych wartosci}
    With comp Do
      With sched Do
      Begin
        initsched(sched,1,nr_comp_proc);
        For i:=1 To nr_comp_proc Do schedul[i]^.nrts:=0;
      End; {with}
    i:=1;

    GetTime(hr,mn,sc,sc100);
    fill_tztzz(tztzz);              {Okreslenie zmodyfikowanych terminow zakonczenia (ztz)}
    sort_tztzz(tztzz, sorted_tztzz, indexed_tab);{Posortowanie zadan w kolejnosci nie rosnacych ztz}

    While non_assigned<>[] Do          {Przydzielenie zadan do procesorow}
    Begin
      With sched Do
      Begin
        execut^:=[];
        For j:=1 To comp.nr_comp_proc Do
        If non_assigned<>[] Then
        Begin
          high_moddate(h, no_exec);
          If h<>0 Then With schedul[j]^ Do {umieszczenie danych o zadaniu w tablicy sched dla procesora j}
          Begin
            a[i].tnr:=h;
            a[i].begint:=t;
            a[i].endt:=t+1;
            a[i].usestime:=1;
            Lateness := a[i].endt - task.task_v[h]^.ddate;
            If (Lateness > MaxLate) Then MaxLate := Lateness;
            end_time:=t+1;
            nrts:=nrts+1;
            non_assigned:=non_assigned-[h];
            execut^:=execut^+[h];
          End; {With}
        End; {If}
        For j:=1 To task.n Do
          If (j in execut^) Then For k:=1 To task.n Do no_exec^[k]:=no_exec^[k]-[j];
        t:=t+1;
        i:=i+1;
      End; {With}
    End; {While}
    sched.c:=t;
    sched.m:=comp.nr_comp_proc;
    sched.Lmax := MaxLate;
    workt:=GetSec(hr,mn,sc,sc100);
    Dispose(execut);
    Dispose(no_exec);
    Dispose(succ);
    Dispose(tztzz);
    Dispose(sorted_tztzz);
    Dispose(indexed_tab);
    is_act_sch:=true;
 {   is_Lmax_crit:=true;}
  End
  Else
  Begin
    MessageBox(0,'Alg 5.20 algorithm doesn`t solve'#10#13
                +'problems with outtree constrains.'#10#13#10#13
                +'No schedule achieved !','WARNING !',mb_OK+mb_IconExclamation);
    is_act_sch:=false;
  End;
  End;  {Alg 5.20}

End.



