{*****************************************************************************}
{*                                                                           *}
{*  Program : Schedul                                                        *}
{*  Unit    : Alg IPL (Information Processing Letters 57 (1996) 35-40        *}
{*                     dr inz. Maciej Drozdowski)                                    *}
{*  This unit contains procedure implementing "IPL" scheduling algorithm.    *}
{*                                                                           *}
{*****************************************************************************}

Unit AlgIPL;

{$C MOVEABLE DEMANDLOAD DISCARDABLE}

Interface

Uses Deklar,WinDos,WinPROCS,WinTypes,Tools;

Procedure AlgIPL_(Var task: task_rec;Var comp: comp_rec; Var sched: schedul_rst;
                   Var workt: real;Var kp: ig);
{ Parameters given:
     task  - task set,
     comp - processor set,
     kp   - number of processors etc.,
  Parameters returned:
     sched   - schedule description,
     workt - how long schedule was computed}

Implementation

Type WordTable = Array [1..MaxN] Of Word;    {tablica numerow zadan}
     PWordTable = ^WordTable;
     ExtendedTable = Array [1..MaxN] Of Extended;
     PExtendedTable = ^ExtendedTable;
     IntegerSet = Set Of 1..255;
     PRk_Set = ^Rk_Set;
     Rk_Set = Record  {Opis zbioru zadan z interwalu rk,rk+1}
              k : Word;
              R : PWordTable;
              number : Word; {liczba zadan w danym zbiorze R}
              next: PRk_Set;
              End;

Procedure AlgIPL_(Var task: task_rec;Var comp: comp_rec; Var sched: schedul_rst;
                   Var workt: real;Var kp: ig);

Var
   i,j,kk,p,l,g,k,f,v,nest,pool,loop: Word;
   z:LongInt;                       {identyfikatory elementow tablic}
   hr,mn,sc,sc100: Word;              {czas}
   t: Extended;                           {aktualny czas wykonywania zadan na procesorach}
   h,wh: Word;                           {ident. przydzielanych zadan}
   XX_Tab : WordTable;
   TT_Tab : PWordTable;
   Delta_Tab : PWordTable;
   Beta_Tab : PExtendedTable;
   Height_Tab :PExtendedTable;
   Release_Times : PExtendedTable;
   ProcTime_Tab : PExtendedTable;
   Rk_List, Pointer, NextPointer : PRk_Set;
   Tau, Tau_Prim : Extended;
   Cmax:Extended;
   DataStatus: Boolean;

{-----------------------------}

Function Tau_bis : Extended;
{Tu obliczamy Tau" wykorzystywane przy obliczniu Tau pkt 2.2.2}
Var i : Word;
    min : Double;
Begin
  min := 21474;
  For i:=1 To task.n Do
    If Beta_Tab^[i] > 0 Then
      If Round(10000*(ProcTime_Tab^[i] / Beta_Tab^[i])) < Round(10000*min) Then
         min := ProcTime_Tab^[i] / Beta_Tab^[i];
  Tau_Bis := min;
End;

{-----------------------------}

Function CheckUpData : Boolean;
Var i: Word;
Begin
  For i:=1 To task.n Do
  Begin
    task.task_v[i]^.tasktime := Round(task.task_v[i]^.tasktime);
    If task.task_v[i]^.tasktime < 1 Then task.task_v[i]^.tasktime :=1;
    task.task_v[i]^.ddate := Round(task.task_v[i]^.ddate) ;
    If task.task_v[i]^.ddate < 1 Then task.task_v[i]^.ddate :=1;
    task.task_v[i]^.rdytime := Round(task.task_v[i]^.rdytime);
    task.task_v[i]^.prcsnum := Round(task.task_v[i]^.prcsnum);
    If task.task_v[i]^.prcsnum < 1 Then task.task_v[i]^.prcsnum :=1;
  End;
  CheckUpData := True;
End;
{-----------------------------}

Function ready_exist(ReadyTime : Extended): Boolean;
{Ta funkcja sprawdza czy release time danego zadania
juz kiedys wystapil czy jeszcze nie - co bedzie wykorzystane
pry ustaleniu liczby interwalow czyli liczby roznych wartosci
release time }
var r:Word;
    exist:Boolean;
Begin
  r:=1;
  exist := False;
  While ((exist = False) And (r<=l)) Do
  Begin
    If ReadyTime = Release_Times^[r] Then exist := True;
    r := r + 1;
  End;
  ready_exist := exist;
End;

{-----------------------------}

Function exist(j : Word; XXX : WordTable): Boolean;
{Ta funkcja sprawdza czy dane zadanie j nalezy do
aktualnie branego pod uwage zbioeu XX }
var r:Word;
    ex:Boolean;
Begin
  r:=1;
  ex := False;
  While ((ex = False) And (r<=task.n)) Do
  Begin
    If (XXX[r] = j) Then ex := True;
    r := r + 1;
  End;
  exist := ex;
End;

{-----------------------------}

Procedure Fill_Height_Tab;
{Inicjujace wypelnienie tabeli z wysokosciami zadan}
Var i:Word;
Begin
  For i:=1 To task.n Do Height_Tab^[i]:= task.task_v[i]^.tasktime / task.task_v[i]^.prcsnum;
End;

{-----------------------------}

Procedure Fill_ProcTime_Tab;
{Inicjujace wypelnienie tabeli pomocniczej czasami przetwarzania zadan - p}
Var i:Word;
Begin
  For i:=1 To task.n Do ProcTime_Tab^[i] := task.task_v[i]^.tasktime;
End;

{-----------------------------}

Procedure Fill_Delta_Tab;
{Inicjujace wypelnienie tabeli liczbami zadanych procesorow}
Var i:Word;
Begin
  For i:=1 To task.n Do Delta_Tab^[i]:= task.task_v[i]^.prcsnum;
End;

{-----------------------------}

Procedure Sort_Rk_Set_By_Height(Var R_Set : PWordTable; num : Word);
{Procedura ta sortuje zadania w zbiorze Rk wedlug wysokosci od najwiekszej do najmniejszej}
var i,j,maks_index : Word;
    maks_height : Extended;
    maks_name : Word;
Begin
  For i := 1 To num-1  Do
  Begin
    maks_height := Height_Tab^[R_Set^[i]];
    maks_name := R_Set^[i];
    maks_index := i;
    For j := (i+1)  To num Do
    Begin
      if Round(10000*Height_Tab^[R_Set^[j]]) > Round(10000*maks_height) Then
      Begin
        maks_height := Height_Tab^[R_Set^[j]];
        maks_name := R_Set^[j];
        maks_index := j;
      End;
    End;
    R_Set^[maks_index] := R_Set^[i];
    R_Set^[i] := maks_name;
  End;
End; {Sort_Rk_Set_By_Height}

{-----------------------------}

Function Sum_Heights_Of_Rk_Set(Var R_Set : PWordTable; num : Word) : Extended;
{Ta procedura oblicza sume wysokosci wszystkich zadan
ze zbioru Rk zeby pozniej sprawdzic czy sa jeszcze jakies
niewykonane zadania (Suma > 0) }
Var
   i : Word;
   Suma : Extended;
Begin
  Suma := 0;
  For i:=1 To num Do If (R_Set^[i] <> 0)
                     Then Suma := Suma + Height_Tab^[R_Set^[i]];
  Sum_Heights_Of_Rk_Set := Suma;
End;

{-----------------------------}

Function Sum_XX_Set_Heights(Var XX_Set : WordTable; num : Word) : Extended;
{Ta procedura oblicza sume wysokosci wszystkich zadan
ze zbioru X zeby pozniej sprawdzic czy sa jeszcze jakies
niewykonane zadania o wysokosci > 0 (Suma > 0) }
Var
   i : Word;
   Suma : Extended;
Begin
  Suma := 0;
  For i:=1 To num Do
  If (XX_Set[i] > 0) Then Suma := Suma + Height_Tab^[XX_Set[i]];
  Sum_XX_Set_Heights := Suma;
End;

{-----------------------------}

Procedure Calculate_Times;
{Procedura obliczajaca Tau' z pkt.2.2.2 algorytmu}
Var j : Word;
    Tau_Temp : Extended;
Begin
  Tau_Prim := 21474;
  For j:=1 To Pointer^.number - 1 Do
    If Round(10000*Height_Tab^[Pointer^.R^[j]]) > Round(10000*Height_Tab^[Pointer^.R^[j+1]]) Then
    Begin
      If Round(10000*(Beta_Tab^[Pointer^.R^[j]]/Delta_Tab^[Pointer^.R^[j]])) <>
         Round(10000*(Beta_Tab^[Pointer^.R^[j+1]]/Delta_Tab^[Pointer^.R^[j+1]]))
      Then
      Begin
        Tau_Temp := (Height_Tab^[Pointer^.R^[j]] - Height_Tab^[Pointer^.R^[j+1]]) /
        (Beta_Tab^[Pointer^.R^[j]]/Delta_Tab^[Pointer^.R^[j]] - Beta_Tab^[Pointer^.R^[j+1]]/Delta_Tab^[Pointer^.R^[j+1]]);
        If Tau_Temp < Tau_Prim Then Tau_Prim := Tau_Temp;
      End;
    End;
End;

{-----------------------------}

Procedure Sort_Release_Times(Var Rel_Tab:PExtendedTable);
{Procedura ta sortuje czasy przybycia od najwczesniejszego do najpozniejszego}
var i,j,actual_index : Word;
    min : Extended;
Begin
  For i := 1 To l-1  Do
  Begin
    min := Rel_Tab^[i];
    actual_index := i;
    For j := (i+1)  To l Do
    Begin
      If Round(10000*Rel_Tab^[j]) < Round(10000*min) Then
      Begin
        min := Rel_Tab^[j];
        actual_index := j;
      End;
    End;
    Rel_Tab^[actual_index] := Rel_Tab^[i];
    Rel_Tab^[i] := min;
  End;
End; {sort_release_times}

{-----------------------------}

Procedure construct_T(Var X: WordTable; Var T: PWordTable; Var suma: Word);
{Procedura ta ze zbioru wejsciowego X wybiera podzbior T}
{zadan o najwyzszych wysokosciach}
var j:Word;
    MaxHeight:Extended;
Begin
  For j:=1 To task.n Do T^[j]:=0;
  MaxHeight:=0;
  suma := 0;
  For j:=1 To task.n Do {znajdujemy maksymalna wysokosc}
  Begin
    If (X[j] > 0)
    Then If (Round(10000*Height_tab^[X[j]]) > Round(10000*MaxHeight))
         Then MaxHeight := Height_tab^[X[j]];
  End;
  For j:=1 To task.n Do {If exist(j,XX_Tab) Then}
    If (X[j] > 0)
    Then If (Round(10000*Height_tab^[X[j]]) = Round(10000*MaxHeight))
         Then If (Round(10000*MaxHeight)/10000) > 0
              Then
              Begin
                T^[j] := X[j];
                suma := suma + task.task_v[X[j]]^.prcsnum;
              End;
End; {construct_T}

{-----------------------------}

Procedure find_capabilities(Rk_in: PWordTable; Var Beta_Tab: PExtendedTable);
{Procedura ta przydziela zadaniom odpowiednia czesc mocy}
{obliczeniowej w zaleznosci od "wysokosci" zadan.       }
var a,b : Extended;
    j, avail : Word;
    X : IntegerSet;
    nazwa,kit, sum_X, sum_delta_T : Word;
Begin
  XX_Tab := Rk_in^;
  For j:= Pointer^.number+1 To task.n Do XX_Tab[j]:= 0;
  {3.1}
  For j:=1 To task.n Do Beta_Tab^[j] := 0;
  avail := sched.m; {avail is the number of free processors}
  {3.2}
  sum_X := 0;
  For j:=1 To task.n Do sum_X := sum_X + XX_Tab[j]; {jesli w XX_Tab sa jakies zadania to suma jest > 0}
  While (avail>0) And
        (sum_X > 0) And
        (Round(10000*Sum_XX_Set_Heights(XX_Tab, Pointer^.number))/10000 > 0)
  Do
  Begin
    {3.2.1}
    construct_T(XX_Tab,TT_tab,sum_delta_T);
    {3.2.2}
    If (sum_delta_T > avail) Then
    Begin
      {3.2.3}
      For j:=1 To task.n Do
      Begin
        nazwa := TT_Tab^[j];
        If nazwa>0 Then Beta_Tab^[nazwa] := (Delta_Tab^[nazwa] * avail) / sum_delta_T;
      End;
      avail := 0;
    End
    Else { Tasks in T can use at most avail processors}
    Begin
      {3.2.4}
      For j:=1 To task.n Do
      Begin
        nazwa := TT_Tab^[j];
        If nazwa>0 Then Beta_Tab^[nazwa] := Delta_Tab^[nazwa];
      End;
      avail := avail - sum_delta_T;
    End;
  {3.3}
  For j:=1 To task.n Do
  Begin
    nazwa:=TT_Tab^[j];
    If nazwa > 0 Then
    Begin
      For kit:=1 To task.n Do If XX_Tab[kit] = nazwa Then XX_Tab[kit] := 0;
    End;  {X := X - T}
  End;
  sum_X := 0;
  For j:=1 To task.n Do sum_X := sum_X + XX_Tab[j]; {jesli w XX_Tab sa jakies zadania to suma jest > 0}
  End; {While loop}
End; {find_capabilities}

{-----------------------------}

Procedure McNaughton;
{Procedura ta szereguje zadania ze zbioru Rk przydzielajc im czas
proporcjonalny do wspolczynnika mocy obliczeniowej (Beta_Tab^[i])
zgodnie z alg. McNaughtona}
Var
  hr,mn,sc,sc100:word;
  i,j,k,np,nt:ig;
  tt: Extended;
  equal,low,great:boolean;
  lasttask,xyz:Word;
  precindex, actualindex : Word;
Begin
  With comp do
  Begin
    Cmax := t + Tau;
    {Przypisanie zadan do procesorow }
    np:=1;
    For xyz:=1 to Pointer^.number Do If Beta_Tab^[Pointer^.R^[xyz]] > 0 Then
    Begin
               nt := Pointer^.R^[xyz];
               If (sched.schedul[np]^.nrts > 0)
               Then If Not(Round(10000*(sched.schedul[np]^.a[sched.schedul[np]^.nrts].endt)) < Round(10000*t))
                 Then
                 Begin
                   low := false; equal := false; great := false;
                   lasttask := sched.schedul[np]^.nrts;
                   low:=(Round(10000*(sched.schedul[np]^.a[lasttask].endt + Tau*Beta_Tab^[nt])) < Round(10000*Cmax));
	           equal:=(Round(10000*(sched.schedul[np]^.a[lasttask].endt + Tau*Beta_Tab^[nt])) = Round(10000*Cmax));
	           great:=(Round(10000*(sched.schedul[np]^.a[lasttask].endt + Tau*Beta_Tab^[nt])) > Round(10000*Cmax));
                 End
                 Else
                 Begin
                   low := false; equal := false; great := false;
                   low:=(Round(10000*(t + Tau*Beta_Tab^[nt])) < Round(10000*Cmax));
                   equal:=(Round(10000*(t + Tau*Beta_Tab^[nt])) = Round(10000*Cmax));
	           great:=(Round(10000*(t + Tau*Beta_Tab^[nt])) > Round(10000*Cmax));
                 End
               Else
               Begin
                 low := false; equal := false; great := false;
                 low:=(Round(10000*(t + Tau*Beta_Tab^[nt])) < Round(10000*Cmax));
	         equal:=(Round(10000*(t + Tau*Beta_Tab^[nt])) = Round(10000*Cmax));
	         great:=(Round(10000*(t + Tau*Beta_Tab^[nt])) > Round(10000*Cmax));
               End;

               If low Then { A task fits in the interval 0..Cmax }
               Begin
                 precindex:=sched.schedul[np]^.nrts;
                 sched.schedul[np]^.nrts:=sched.schedul[np]^.nrts+1;
                 actualindex:=sched.schedul[np]^.nrts;
	         If (actualindex>1)
                 Then If Not (Round(10000*(sched.schedul[np]^.a[precindex].endt)) < Round(10000*t) )
                   Then sched.schedul[np]^.a[actualindex].begint:=
                        sched.schedul[np]^.a[precindex].endt
                   Else sched.schedul[np]^.a[actualindex].begint:=t
	         Else sched.schedul[np]^.a[actualindex].begint:=t;
	         sched.schedul[np]^.a[actualindex].tnr:=nt;
	         sched.schedul[np]^.a[actualindex].endt:=
                 sched.schedul[np]^.a[actualindex].begint + Tau*Beta_Tab^[nt];
	         sched.schedul[np]^.a[actualindex].usestime := Tau*Beta_Tab^[nt];
               End;

	       If equal then {The tasks fill completely 0..Cmax interval }
	       Begin
                 precindex:=sched.schedul[np]^.nrts;
                 sched.schedul[np]^.nrts:=sched.schedul[np]^.nrts+1;
                 actualindex:=sched.schedul[np]^.nrts;
	         If (actualindex>1)
                 Then If Not(Round(10000*sched.schedul[np]^.a[precindex].endt) < Round(10000*t))
                   Then sched.schedul[np]^.a[actualindex].begint:=sched.schedul[np]^.a[precindex].endt
                   Else sched.schedul[np]^.a[actualindex].begint:=t
	         Else sched.schedul[np]^.a[actualindex].begint:=t;
	         sched.schedul[np]^.a[actualindex].tnr := nt;
                 sched.schedul[np]^.a[actualindex].endt := Cmax;
		 sched.schedul[np]^.a[actualindex].usestime := Tau*Beta_Tab^[nt];
                 sched.schedul[np]^.end_time:=Cmax;
		 np:=np+1;
	       End;

               If great Then {The task doesn't fit into 0..Cmax interval}
               Begin
                 tt := Tau*Beta_Tab^[nt];
                 While great Do
                 Begin
                   precindex:=sched.schedul[np]^.nrts;
                   sched.schedul[np]^.nrts:=sched.schedul[np]^.nrts+1;
                   actualindex:=sched.schedul[np]^.nrts;
	           If (actualindex>1)
                   Then If Not(Round(10000*sched.schedul[np]^.a[precindex].endt) < Round(10000*t))
                     Then sched.schedul[np]^.a[actualindex].begint :=sched.schedul[np]^.a[precindex].endt
                     Else sched.schedul[np]^.a[actualindex].begint:=t
	           Else sched.schedul[np]^.a[actualindex].begint:=t;
	           sched.schedul[np]^.a[actualindex].tnr:=nt;
                   sched.schedul[np]^.a[actualindex].endt:=Cmax;
		   sched.schedul[np]^.a[actualindex].usestime:=Cmax-sched.schedul[np]^.a[sched.schedul[np]^.nrts].begint;
		   tt:=tt-sched.schedul[np]^.a[actualindex].usestime;
                   sched.schedul[np]^.end_time:=Cmax;
                   np:=np+1;
                   actualindex := sched.schedul[np]^.nrts;
                   low := false; equal := false; great := false;
                   If (actualindex > 0)
                   Then If Not(Round(10000*sched.schedul[np]^.a[actualindex].endt) < Round(10000*t))
                     Then great:=(Round(10000*(sched.schedul[np]^.a[actualindex].endt + tt)) > Round(10000*Cmax))
                     Else great := ( Round(10000*(t + tt)) > Round(10000*Cmax))
                   Else great := ( Round(10000*(t + tt)) > Round(10000*Cmax));
                 End;
                 low := false; equal := false; great := false;
                 actualindex := sched.schedul[np]^.nrts;
                 If (actualindex > 0)
                 Then If Not(Round(10000*sched.schedul[np]^.a[actualindex].endt) < Round(10000*t))
                   Then
                   Begin
                     low := Round(10000*(sched.schedul[np]^.a[actualindex].endt + tt)) < Round(10000*Cmax);
                     equal:= Round(10000*(sched.schedul[np]^.a[actualindex].endt + tt)) = Round(10000*Cmax);
                   End
                   Else
                   Begin
                     low:= Round(10000*(t + tt)) < Round(10000*Cmax);
                     equal:= Round(10000*(t + tt)) = Round(10000*Cmax);
                   End
                 Else
                 Begin
                   low:= Round(10000*(t + tt)) < Round(10000*Cmax);
                   equal:= Round(10000*(t + tt)) = Round(10000*Cmax);
                 End;

                 If low Then { A task fits in the interval 0..Cmax }
                 Begin
                   precindex:= sched.schedul[np]^.nrts;
                   sched.schedul[np]^.nrts:=sched.schedul[np]^.nrts+1;
                   actualindex:= sched.schedul[np]^.nrts;
	           If (actualindex>1)
                   Then If Not (Round(10000*sched.schedul[np]^.a[precindex].endt) < Round(10000*t))
                     Then sched.schedul[np]^.a[actualindex].begint:=
                          sched.schedul[np]^.a[precindex].endt
                     Else sched.schedul[np]^.a[actualindex].begint:=t
	           Else sched.schedul[np]^.a[actualindex].begint:=t;
	           sched.schedul[np]^.a[actualindex].tnr:=nt;
	           sched.schedul[np]^.a[actualindex].endt:=sched.schedul[np]^.a[sched.schedul[np]^.nrts].begint+tt;
	           sched.schedul[np]^.a[actualindex].usestime:=tt;
                 End;

                 If equal then {The tasks fill completely 0..Cmax interval }
	         Begin
                   precindex:=sched.schedul[np]^.nrts;
                   sched.schedul[np]^.nrts:=sched.schedul[np]^.nrts+1;
                   actualindex:=sched.schedul[np]^.nrts;
	           If (actualindex>1)
                   Then If Not(Round(10000*sched.schedul[np]^.a[precindex].endt) < Round(10000*t))
                     Then sched.schedul[np]^.a[actualindex].begint:=
                          sched.schedul[np]^.a[precindex].endt
                     Else sched.schedul[np]^.a[actualindex].begint:=t
	           Else sched.schedul[np]^.a[actualindex].begint:=t;
	           sched.schedul[np]^.a[actualindex].tnr:=nt;
                   sched.schedul[np]^.a[actualindex].endt:=Cmax;
		   sched.schedul[np]^.a[actualindex].usestime:=tt;
                   sched.schedul[np]^.end_time:=Cmax;
		   np:=np+1;
	         End;
               End;   {IF GREAT}
    End;  { For nt =1 to }
  End; {with comp do}
{  is_act_sch:=true;}
End;

{-----------------------------}

Begin               {Alg IPL}
    GetTime(hr,mn,sc,sc100);
    memory:=0;
    New(TT_Tab);
    New(Height_tab);
    New(Beta_Tab);
    New(Delta_Tab);
    New(Release_Times);
    New(ProcTime_Tab);
    {Ustawienie poczatkowych wartosci}
    With comp Do
      With sched Do
      Begin
        initsched(sched,1,nr_comp_proc);
        For i:=1 To nr_comp_proc Do schedul[i]^.nrts:=0;
      End; {with}
    i:=1;
    sched.m := comp.nr_comp_proc;
    {1}
    DataStatus := CheckUpData;
    Fill_Height_Tab;
    Fill_Delta_Tab;
    Fill_ProcTime_Tab;
    l:=0;
    For p:=1 To task.n Do Release_Times^[p]:= -1;
    For p:=1 To task.n Do {We'll find out how many different values of release times is present}
    Begin
      IF ready_exist(task.task_v[p]^.rdytime) = False Then
      Begin
        l := l+1;
        release_times^[l] := task.task_v[p]^.rdytime ;
      End;
    End; {l - number of different values of release times}
         {so l is number of intervals}
         {release_times^ - table with these values}
    nest := l;
    Sort_Release_Times(Release_Times);
    Release_Times^[l+1] := 21474;
    t := 0;
    t := Release_Times^[1];
{    Time_Difference := Release_Times^[1] - 0;}
{    For p:=1 To task.n      }
    New(Rk_List);
    New(Pointer);
    Pointer := Rk_List;
    For g:=1 To l Do
    Begin
      Pointer^.k := g;
      Pointer^.number := 0;
      New(Pointer^.R);
      For z:=1 To MaxN Do Pointer^.R^[z] := 0;
      For z:=1 To task.n Do
        If Round(10000*task.task_v[z]^.rdytime) = Round(10000*Release_Times^[g]) Then
        Begin
          Pointer^.R^[Pointer^.number+1]:= z;
          Pointer^.number := Pointer^.number + 1;
        End;
      Sort_Rk_Set_By_Height(Pointer^.R, Pointer^.number);
      If (g < l) Then New(Pointer^.next)
      Else Pointer^.next := NIL;
      Pointer := Pointer^.next
    End;
    For i:=1 To l Do memory := memory + SizeOf(RK_List^);
    memory:=memory+SizeOf(TT_Tab^)+ SizeOf(Height_Tab^) + SizeOf(ProcTime_Tab^) +
            + SizeOf(Beta_Tab^)+SizeOf(Delta_Tab^)+SizeOf(Release_Times^);
    Pointer := Rk_List; {wskazujemy na pierwszy element}
    {Powyzej utworzylismy liste zbiorow Rk,wypelnilismy te zbiory
    {zadaniami i posortowalismy je wedlug wysokosci}
    {2}
    For k:=1 To l Do
    Begin
      {2.1}
      Sort_Rk_Set_By_Height(Pointer^.R, Pointer^.number);
      {2.2}
      wh := 0;
      While (Round(10000*Release_Times^[k+1]) > Round(10000*t)) And
      (Round(10000*Sum_Heights_Of_Rk_Set(Pointer^.R, Pointer^.number))/10000  > 0) Do
      Begin
        wh := wh + 1;
        {2.2.1}
        Find_Capabilities(Pointer^.R, Beta_Tab);
        {2.2.2}
        Calculate_Times;
        {2.2.3}
        Tau := Tau_Prim;
        If Round(10000*Tau_Bis) < Round(10000*Tau) Then Tau := Tau_Bis;
        If Round(10000*(Release_Times^[k+1]-t)) < Round(10000*Tau) Then Tau := (Release_Times^[k+1] - t);
        {2.2.4}
        McNaughton;
        {2.2.5}
        For f:=1 To Pointer^.number Do
        Begin
          Height_Tab^[Pointer^.R^[f]] := Height_Tab^[Pointer^.R^[f]] -
          Tau * (Beta_Tab^[Pointer^.R^[f]] / Delta_Tab^[Pointer^.R^[f]]);
          If Height_Tab^[Pointer^.R^[f]] < 0 Then Height_Tab^[Pointer^.R^[f]] := 0;
          ProcTime_Tab^[Pointer^.R^[f]] := ProcTime_Tab^[Pointer^.R^[f]] -
          Tau * Beta_Tab^[Pointer^.R^[f]];
          If ProcTime_Tab^[Pointer^.R^[f]] < 0 Then ProcTime_Tab^[Pointer^.R^[f]] := 0;
        End;
        {2.2.6}
        t := t + Tau;
      End; {While loop}
      {2.3}
      If Round(10000*Sum_Heights_Of_Rk_Set(Pointer^.R, Pointer^.number)) / 10000 > 0 Then
      Begin
        NextPointer := Pointer^.next;
        For v:=1 To Pointer^.number Do
        Begin
          If Round(10000*Height_Tab^[Pointer^.R^[v]])/10000 > 0 Then
          Begin
            NextPointer^.R^[NextPointer^.number + 1] := Pointer^.R^[v];
            NextPointer^.number := NextPointer^.number + 1;
          End;
        End;
      End
      Else
      Begin
        If (k < l)
        Then If (Round(10000*Release_Times^[k+1]) > Round(10000*t))
             Then t := Release_Times^[k+1];
      End;
    Pointer := Pointer^.next;
    End; {For loop}
    sched.c:=t;
    sched.m:=comp.nr_comp_proc;
{    sched.Lmax := max_lateness;}
    workt:=GetSec(hr,mn,sc,sc100);
    Pointer := Rk_List;
    For loop:=1 To nest-1 Do Pointer := Pointer^.next;
    Dispose (Pointer^.R);
    nest := nest -1;
    For pool := l-1 DownTo 1 Do
    Begin
      Pointer := Rk_List;
      For loop:=1 To nest-1 Do Pointer := Pointer^.next;
      Dispose (Pointer^.next);
      Dispose (Pointer^.R);
      nest := nest -1;
    End;
    Dispose(Pointer);
    Dispose(TT_Tab);
    Dispose(Height_tab);
    Dispose(Beta_Tab);
    Dispose(Delta_Tab);
    Dispose(Release_Times);
    Dispose(ProcTime_Tab);
    is_act_sch:=true;
  End;
End.  {Alg IPL}