{*****************************************************************************}
{*                                                                           *}
{*                                                                           *}
{*  Program : Simul                                                          *}
{*  Unit    : CoGrah                                                         *}
{*                                                                           *}
{*  This unit contains procedure implementing Coffman-Graham's               *}
{*  scheduling algorithm.                                                    *}
{*                                                                           *}
{*****************************************************************************}

Unit CoGrah;

{$C MOVEABLE DEMANDLOAD DISCARDABLE}

Interface

Uses Deklar,WinDos,WinPROCS,WinTypes,Tools,ToolHelp;

Procedure CoffGrah(Var task: task_rec;Var comp: comp_rec; Var sched: schedul_rst;
                   Var workt: real;Var kp: ig);
{ Parameters given:
     task  - task set,
     comp - processor set,
     kp   - number of processors etc.,
  Parameters returned:
     sched   - schedule description,
     workt - how long schedule was computed}

Implementation

Type Tetyk = Array [0..MaxN] Of Byte;     {tablica etykiet dla zadan}
     Petyk = ^Tetyk;
     Tsucc = Array [1..MaxN] Of  fs;      {tablica zbiorow nastepnikow zadan}
     Psucc = ^Tsucc;
     Tno_exec = Array [1..MaxN] Of fs;    {tablica zbiorow nie wykonanych poprzednikow}
     Pno_exec = ^Tno_exec;

Procedure CoffGrah(Var task: task_rec;Var comp: comp_rec; Var sched: schedul_rst;
                   Var workt: real;Var kp: ig);

Var
   i,j,q: Byte;                       {identyfikatory elementow tablic}
   etyk: Petyk;
   hr,mn,sc,sc100: Word;              {czas}
   non_assigned: fs;                  {zbior zadan nieprzydzielonych}
   succ: Psucc;
   no_exec: Pno_exec;
   wsk: ptrprem;
   t: Byte;                           {aktualny czas wykonywania zadan na obu maszynach}
   g,h: Byte;                         {zmienne pomocnicze- ident. zadan}


Procedure Labels(n: Byte; Var etyk:Petyk; Var succ:Psucc);
{This procedure assigns labels to tasks.}
Var
    et,i,j: Byte;
    found: Boolean;
Begin
  For i:=1 To n Do etyk^[i]:=0;
  etyk^[0]:=0;
  et:=1;
  While et<=n Do
    For i:=1 To n Do
    Begin
      If ((succ^[i]=[]) And (etyk^[i]=0)) Then Begin
                                               etyk^[i]:=et;
                                               et:=et+1;
                                               For j:=1 To n Do
                                                 succ^[j]:=succ^[j]-[i];
                                             End; {If}
    End;  {For}
End;  {Labels}


Procedure najw(Var x: Byte; Var etyk:Petyk; Var no_exec:Pno_exec);
{Procedura ta zwraca numer zadania, ktore ma najwyzsza etykiete,}
{nalezy do zbioru nie przydzielonych zadan i ktorego wszystkie  }
{poprzedniki zostaly wykonane.                                  }
Var
   naj,q: Byte;
Begin
  naj:=0;
  For q:=1 To task.n Do
    If (q in non_assigned) And (etyk^[q]>etyk^[naj]) And (no_exec^[q]=[]) Then naj:=q;
  x:=naj;
End; {najw}



Begin
  memory:=0;
  New(etyk);
  New(succ);
  New(no_exec);
  if comp.nr_comp_proc>2 Then
  Begin
   MessageBox(0,'Number of processors exceeds 2.'#10#13
               +'Because Coffman-Grahams alghorithm runs on two processors,'#10#13
               +'now the number of processors is restricted to 2.','WARRNING !',mb_OK+mb_IconInformation);
   For i:=3 To comp.nr_comp_proc Do Dispose(comp.comp_pr[i]);
   comp.nr_comp_proc:=2;         {Ograniczenie zbioru procesorow do 2; takze w linii powyzej}
  End;
  non_assigned:=[];             {tworzenie zbioru non_assigned i tablicy no_exec}
  For i:=1 To task.n Do no_exec^[i]:=[];
  For i:=1 To task.n Do
  Begin
    succ^[i]:=[];
    non_assigned:=non_assigned+[i];
    If task.task_v[i]^.next<>nil Then
    Begin
      wsk:=task.task_v[i]^.next;
      While wsk<>nil Do
      Begin
        no_exec^[wsk^.nrprem]:=no_exec^[wsk^.nrprem]+[i];
                                        {wpis do tab.no_exec dla 'wsk^.nrprem' zadania 'i' jako poprzednika}
        succ^[i]:=succ^[i]+[wsk^.nrprem];                   {wpis do tablicy nastepnikow}
        wsk:=wsk^.next;                                   {wybranie nastepnego elementu z listy nastepnikow}
      End; {While}
    End; {If}
  End; {For}

  For i:=1 To task.n Do memory:=memory+SizeOf(no_exec^[i])+SizeOf(succ^[i]);
  memory:=memory+SizeOf(etyk^)+SizeOf(non_assigned);

  t:=0; i:=1; j:=1;     {ustawienie poczatkowych wartosci}
  With sched Do
  Begin
    initsched(sched,1,2);
    schedul[1]^.nrts:=0;
    schedul[2]^.nrts:=0;
  End; {With}

  GetTime(hr,mn,sc,sc100);
  Labels(task.n,etyk,succ);          {uruchomienie procedury nadawania etykiet zadaniom}
  While non_assigned<>[] Do         {przydzielanie zadan do procesorow}
  Begin
    With sched Do                     {wpisywanie uszeregowania do tablicy wynikowej}
    Begin
      najw(h,etyk,no_exec);
      If etyk^[h]<>0 Then Begin
                            With schedul[1]^ Do      {umieszczenie danych o zadaniu w tablicy sched dla procesora M1}
                            Begin
                              a[i].tnr:=h;
                              a[i].begint:=t;
                              a[i].endt:=t+1;
                              a[i].usestime:=1;
                              end_time:=t+1;
                              nrts:=nrts+1;
                            End; {With}
                            i:=i+1;
                            non_assigned:=non_assigned-[h];               {usuniecie zadania k ze zbioru nie wykonanych zadan}
                            End; {If}
      najw(g,etyk,no_exec);
      If etyk^[g]<>0 Then Begin
                            With schedul[2]^ Do      {umieszczenie danych o zadaniu w tablicy sched dla procesora M2}
                            Begin
                              a[j].tnr:=g;
                              a[j].begint:=t;
                              a[j].endt:=t+1;
                              a[j].usestime:=1;
                              end_time:=t+1;
                              nrts:=nrts+1;
                            End; {With}
                            j:=j+1;
                            non_assigned:=non_assigned-[g];               {usuniecie zadania k ze zbioru nie wykonanych zadan}
                            End; {If}
      For q:=1 To task.n Do Begin
                             no_exec^[q]:=no_exec^[q]-[h];  {usuniecie h ze zb. nie wykonanych poprzedn.}
                             no_exec^[q]:=no_exec^[q]-[g];  {usuniecie g ze zb. nie wykonanych poprzedn.}
                           End; {For}
    End; {With}
    t:=t+1;
  End; {While}
  sched.c:=t;
  sched.m:=2;
  workt:=GetSec(hr,mn,sc,sc100);
  is_act_sch:=true;
  Dispose(no_exec);
  Dispose(succ);
  Dispose(etyk);
End; {CoffGrah}

End.
