{$I PCoptions}
{$C MOVEABLE DEMANDLOAD}

unit Deklar;

{*****************************************************************************}
{*                                                                           *}
{*  Program : Simul                                                          *}
{*  Unit    : Deklar                                                         *}
{*                                                                           *}
{*  This Unit contains definitions of variable types used to describe:       *}
{*    sets of tasks and processors;                                          *}
{*    method of data generation;                                             *}
{*    the schedule.                                                          *}
{*  This unit contains declarations of common procedures used by all         *}
{*  other units.                                                             *}
{*                                                                           *}
{*****************************************************************************}

interface

const  MaxM=200{550};                   { Max. number of processors}
       MaxN=500;                   { Max. number of tasks}
       max_comp=15;                { Max. number of real processors in one composite processor (multiprocessors)}
       max_compr=3200;              { Max. number of different tasks (or parts) on one processor}
       re=0.000001;                { Accuracy of computations}

type   Strg=array[0..80] of Char;
       lig=longint;
       ig=integer;
       fs=set of 1..255;
       succtab=array[1..MaxN] of ig;     { auxiliary arry for a generation of precedence constraints}

      {***************     T A S K S    ******************}

       ptrprem=^prem;
       prem=record                 { Element of tasks precedence constraint list}
                  nrprem:ig;       { Number of task succeeding current task}
                  next:ptrprem;    { Pointer to the next task on the list of precedence}
            end;
      task_ptr=^task_rec0;         { Pointer to task record }
      task_rec0=record             { Task Record }
                nr:ig;             { Task number}
                tasktime:real;     { Execution time}
                rdytime:real;      { Ready time }
                ddate:real;        { Due date }
                weight:ig;         { Priority weight}
                prcsnum:ig;        { Number of necessary processors, for multiprocessor tasks}
                dedication:fs;     { Set of processors required by given task (if a task is to
                                     be executed on more than one processor in consecutive order
                                     eg. as in some kind of shop then you have to model it
                                     with several tasks and precedence constraints}
                next:ptrprem;      { Pointer to the first succeding task (precedence constraint)}

{ The User himself may add (remove) here any necessary (unnecesary) task parameter
 (or change its type) depending on needs and machine abilities.
 Example for a flow,open,job shop
    operations:array [1..MaxM] of real;
 operation[i] is processing time on the processor Pi,0 means not processed on Pi}

              end;

      task_arr=array[1..MaxN] of task_ptr;      { Tasks pointers array}
      task_rec=record                           { Full tasks set description record}
                     n:ig;                      { number of tasks}
                     nr_k:ig;                   { number of k-processor tasks}
                     task_v:task_arr;           { Tasks array}
                end;


      {***************     P R O C E S S O R S       ******************}

      comp_el=record                     { Component of one composite processor}
                pr_nr:ig;                { original processor number }
                comp_speed:ig;           { original processor speed }
                beg_time:real;           { original processor work start time in the component of composite processor}
                end_time:real;           { original processor work break time in the component of composite processor}
                comp_pc:real;            { Processing capacity of the component=original_processor_speed*(end_time-beg_time)}
              end;

      comp_proc=record                       { Composite processor (multiprocesor tasks scheduling)}
                   proc_nr:    ig;           { Processor number }
                   speed  :    ig;           { Processor speed }
                   pc     :    real;         { Processing capacity}
                   nr_comp:    ig;           { number of real processors in the composite processor}
                   comp_a :    array[1..max_comp]of comp_el;  { Components array }
                end;

      comp_proc_ptr=^comp_proc;                  { Pointer to composite processor record}
      comp_procs=array[1..MaxM]of comp_proc_ptr; { Composite processor array}
      comp_rec=record                            { Record fully describing processor set (for real and composite processors)}
                 nr_comp_proc:ig;                { Number of processors}
                 comp_pr:comp_procs;             { Composite processor pointers array}
               end;

      {****************  S C H E D U L E ******************}

      task_rec1=record          { Task (or a part of task) processed on a given processor}
                  tnr:ig;       { Task number}
                  begint:real;  { Execution start time}
                  endt:real;    { Execution break time}
                  usestime:real;{ Execution time}
                end;

      mresult=record            { Decription of the processor after schedule}
                end_time:real;  { Processor stop time}
                pr_speed:ig;    { Processor speed}
                nrts:ig;        { Number of different tasks (or parts of tasks) executed by the processor}
                a:array[1..max_compr] of task_rec1;{ Array of tasks (or parts of tasks) executed by the processor}
              end;

      mresult_ptr=^mresult;     { Pointer to a record describing schedule on one processor }
      schedul_rst=record        { Full schedule record}
                c:real;         { Cmax }
             lmax:real;         { Lmax }
                m:ig;           { number of processors}
                schedul:array[1..maxM] of mresult_ptr;{ Array of processors after schedule}
              end;


      {************* CONFIGURATION OF DATA GENERATION ******************}


    gendatatype=(data,rnd1,rndnorm,rndexp,cons);  { types of data sources:
                  data - constant values given explicitly from keyboard or screen,
                  rnd1 - uniform random generator (flat distribution),
                  rndnorm - normal distribution generator,
                  rndexp  - exponential distribution generator,
                  cons - constant value }

    succdata=(datas,nosucc,chain,intree,outtree);  { types of precedence constraint graph }
    TypeF=(fi,mn);

    gendata=record                        { complete description of one value generation metho}
            form:gendatatype;             { method of value generation}
            par1,par2:real;               { parameters for generators}
           end;
    generate=record                       { complete description of the data set deneration for tasks and processors}
          speed_g:gendata;                { processors speeds }
                                          { tasks parameters }
          task_p,task_r,task_d,           { processing, ready times due-date}
          task_k,task_ded:gendata;        { # of processors req., processors req.}
          succ_g:succdata;                { precedence type}
         end;

    Exp=record
         E,DecE: Integer;
        end;
    All=record
         T: task_rec;
         C: comp_rec;
         G: generate;
        end;
    HowFile = (add,newF,over);

    SetupSaveInf=record  {Setup Save Information Box record}
           speed: Word;
           exect: Word;
           ready: Word;
           ddate: Word;
           numpr: Word;
           Indce: Word;
           Prec : Word;
           ofile: String;
          end;



var
       comp:comp_rec;          { Processor set description}
       sched:schedul_rst;      { Schedule description}
       tasks:task_rec;         { Task set description}
       kp:integer;             { Max number of processors required by a multiprocessor task}
       memory: Longint;
       is_act_sch:boolean;     { Is any schedule ready? flag}
       is_act_data:boolean;    { Is any data set ready? flag}
       is_act_succ:boolean;    { Is any succeding ready? flag}
       is_act_task:boolean;    { Is any task ready? flag}
       is_act_proc:boolean;    { Is any processors ready? flag}
       is_act_descr:boolean;   { Is any correct generate structure ready? flag}
       is_Lmax_crit:boolean;   { Criterion flag }
       ExperInProgress:Boolean;
       gen:generate;           { Configuration description}
       criterion:string;       { Optymality criterion name}
       workt: Real;
       Experiment: Exp;
       EverageCmax:Real;
       EverageExecTime:Real;
       FileInEdit: Boolean;
       FirstExper: Boolean;
       PosX,PosY:Integer;
       Log: Boolean;
       fullform:Boolean;
       HuConstrain: succdata;
       LPName: String[12];                  {LP-solver name}
       LPPath,WorkDir: String[80];          {Path to LP-solver, working directory}
       WindowsDir: Array [0..80] Of Char;   {windows directory}
implementation
end.
