{*****************************************************************************}
{*                                                                           *}
{*  Program : Simul                                                          *}
{*  Unit    : GonzSahni                                                      *}
{*                                                                           *}
{*  This unit contains procedure implementing Gonzalel&Sahni's               *}
{*  scheduling algorithm.                                                    *}
{*                                                                           *}
{*****************************************************************************}

unit GonSahni;

{$C MOVEABLE DEMANDLOAD DISCARDABLE}

interface

uses WinProcs,WinDos,Deklar,Tools,ToolHelp;

procedure Gonzalez(tasks:task_rec; comp:comp_rec; var sched:schedul_rst;
                   var workt:real; kp:ig);
{ Parametrs given:
     task - tasks set,
     comprec - processor set,
     kp   - number of processors executing widest multiprocessor task,
  Parametrs returned:
     sched - schedule description,
     workt - how long schedule was computed}

implementation

procedure Gonzalez(tasks:task_rec; comp:comp_rec; var sched:schedul_rst;var workt:real;kp:ig);
var
  tasksD:task_rec;
  comprec:comp_rec;
  T:Task_rec0;
  ext,Cmax:real;
  i,K,K1:Integer;
  hr,mn,sc,sc100:word;
  DevT:Real;
  Dum:Boolean;
{  i,j,k,np,nt:ig;
  tt: real;
  ch:char;
  equal,low,great:boolean;}

procedure InitDuplik(var T:task_rec;var C:comp_rec);
var
i:Integer;
begin
 For i:=1 to Comp.nr_comp_proc do
 begin
  new(C.comp_pr[i]);
  C.comp_pr[i]^:=Comp.Comp_pr[i]^;
 end;
 C.nr_comp_proc:=Comp.nr_comp_proc;
 For i:=1 to tasks.n do
 begin
  new(T.task_v[i]);
  T.task_v[i]^:=tasks.task_v[i]^;
 end;
 T.n:=tasks.n;
 T.nr_k:=tasks.nr_k;
end;



procedure SetComp(var C:Comp_el;prnr,sp:integer;beg,endT,pc:real);
{this function sets given record of comp_el type}
begin
  with C do
  begin
   pr_nr:=prnr;
   comp_speed:=sp;
   beg_time:=beg;
   end_time:=endT;
   comp_pc:=pc
  end;
end;

procedure DelTask(t:task_rec0);
var
 i:Integer;
begin
 with tasksd do
 begin
  if n=1 then
  begin
   {dispose(task_v[1]);}
   n:=0;
  end else
  begin
   for i:=2 to n do
   task_v[i-1]^:=task_v[i]^;
  { dispose(task_v[n]);}
   dec(n);
  end;
 end;
end;


procedure SetProc(var Proc:comp_proc;prnr,sp:Integer;p:real;nrproc:Integer;ca:array of comp_el);
var
 i:Integer;
begin
  with Proc do
  begin
   proc_nr:=prnr;
   speed:=sp;
   pc:=p;
   nr_comp:=nrproc;
   for i:=1 to max_comp do
    Comp_a[i]:=ca[i];
  end;
end;

procedure SetDummy(var C:comp_proc);
begin
 with C do
 begin
  proc_nr:=0;
  speed:=0;
  pc:=0;
  nr_comp:=1;
  SetComp(comp_a[1],0,0,0,Cmax,0);
 end;
end;

procedure SetProcRec(var Cr:Comp_rec;Comp1:Comp_rec;m:Integer);
var
 i:Integer;
begin
  for i:=1 to m do
   begin
     new(cr.comp_pr[i]);
     cr.comp_pr[i]^:=comp1.comp_pr[i]^;
   end;
end;

function FindCmax(tasksd:task_rec;comp1:comp_rec):Real;
{ This function finds Cmax using the following equation:
                           k       k       n       m
  Cmax>= max (   max    (  E  pj / E  si), E  pj / E  Si )
              1<=k<=m-1   j=1     i=1     j=1     i=1
}
var
  ECap,MCap,SumSpeed,SumTTime:Real;
  i,j,k:Integer;
begin
  SumSpeed:=0;
  SumTTime:=0;
  MCap:=0;
  with comp1 do
   for i:=1 to nr_comp_proc do
    SumSpeed:=SumSpeed+comp_pr[i]^.speed;
  with tasksd do
   for j:=1 to n do
    SumTTime:=SumTTime+task_v[j]^.tasktime;
  ECap:=SumTTime/SumSpeed;
  SumSpeed:=0;
  SumTTime:=0;
  for k:=1 to comp1.nr_comp_proc-1 do
  begin
   with comp1 do
    for i:=1 to k do
     SumSpeed:=SumSpeed+comp_pr[i]^.speed;
   with tasksd do
    for j:=1 to k do
     SumTTime:=SumTTime+task_v[j]^.tasktime;
   if MCap<=SumTTime/SumSpeed then MCap:=SumTTime/SumSpeed;
  end;
  if MCap>ECap then FindCmax:=MCap else FindCmax:=ECap;
end;

procedure SortProc(var comp1:comp_rec);
var
 i,j:integer;
 p:comp_proc_ptr;
begin
  with comp1 do
  begin
   for i:=2 to nr_comp_proc do
    for j:=nr_comp_proc downto i do
    if comp_pr[j-1]^.comp_a[1].comp_speed<
       comp_pr[j]^.comp_a[1].comp_speed then
    begin
     p:=comp_pr[j-1];
     comp_pr[j-1]:=comp_pr[j];
     comp_pr[j]:=p;
    end;
  end;
end;

function OneNotDone(c:comp_rec):boolean;
var
 i,k:Integer;
begin
  k:=0;
  for i:=1 to C.nr_comp_proc do
  if C.comp_pr[i]^.pc<>0 then k:=k+1;
  if k=1 then OneNotDone:=True else OneNotDone:=False;
end;

procedure SortProcC(var comp1:comp_rec);
var
 i,j:integer;
 p:comp_proc_ptr;
begin
  with comp1 do
  begin
   for i:=2 to nr_comp_proc do
    for j:=nr_comp_proc downto i do
    if comp_pr[j-1]^.pc<comp_pr[j]^.pc then
    begin
     p:=comp_pr[j-1];
     comp_pr[j-1]:=comp_pr[j];
     comp_pr[j]:=p;
    end;
  end;
end;

procedure SortTask(var tasksd:task_rec);
var
 i,j:integer;
 p:task_ptr;
begin
  with tasksd do
  begin
   for i:=2 to n do
    for j:=n downto i do
    if task_v[j-1]^.tasktime<task_v[j]^.tasktime then
    begin
     p:=task_v[j-1];
     task_v[j-1]:=task_v[j];
     task_v[j]:=p;
    end;
  end;
end;

function CheckBound(LB0,UB0:real;C1:Comp_proc):Boolean;
{this function checks if the given bound is within the working time
 of  real processor in the composite processor}
var
i:Integer;
begin
 CheckBound:=false;
 for i:=1 to C1.nr_comp do
  if (C1.comp_a[i].beg_time<=LB0)and (C1.comp_a[i].end_time>=UB0)then
  CheckBound:=true;
end;

function CheckT(time:real;T:task_rec0;C1,C2:Comp_proc):Integer;
{this function checks if given time divides working time of machines C1, C2
 in such way that capacity of machine C1 in [0,time] interval plus capacity
 of machine C2 in [time,Cmax] interval is:
            - equal           value(0)
            - greater         value(1)
            - smaller         value(-1)
 then the task T execution time}
 var
   i:Integer;
   cap1,cap2:Real;
 begin
   i:=1;
   Cap1:=0;
   while C1.comp_a[i].end_time<=time do
   begin
    Cap1:=Cap1+C1.comp_a[i].comp_pc;
    i:=i+1;
   end;
   if C1.comp_a[i].end_time>time then
    Cap1:=Cap1+C1.comp_a[i].comp_speed*(time-C1.comp_a[i].beg_time);
   i:=1;
   Cap2:=0;
   while C2.comp_a[i].end_time<=time do i:=i+1;
   if C2.comp_a[i].end_time>time then
   begin
    Cap2:=Cap2+C2.comp_a[i].comp_speed*(C2.comp_a[i].end_time-time);
    i:=i+1;
   end;
   while i<=C2.nr_comp do
   begin
    Cap2:=Cap2+C2.comp_a[i].comp_pc;
    i:=i+1;
   end;
   if T.tasktime>Cap1+Cap2 then CheckT:=1;
   if T.tasktime=Cap1+Cap2 then CheckT:=0;
   if T.tasktime<Cap1+Cap2 then CheckT:=-1;
 end;

procedure FindBound(var LB0,UB0:Real;T:task_rec0;C1,C2:comp_proc);
var
 TempTime:Real;
begin
 If not CheckBound(LB0,UB0,C1)or not CheckBound(LB0,UB0,C2)then
   begin
    TempTime:=(UB0-LB0)/2+LB0;
    if CheckT(TempTime,T,C1,C2)=1 then LB0:=TempTime;
    if CheckT(TempTime,T,C1,C2)=-1 then UB0:=TempTime;
    if CheckT(TempTime,T,C1,C2)<>0 then FindBound(LB0,UB0,T,C1,C2);
   end;
end;

function FindT(T:task_rec0;C1,C2:comp_proc):Real;
{this function returns Time which devides working time of machines C1, C2
 in such way that capacity of machine C1 in the interval [0,Time] plus
 capacity of machine C2 in the interval [Time,Cmax] equals the execution
 time of task T}
 var
  i,j,ind:Integer;
  LB,UB:Real;
  Cap1,Cap2:Real;
 begin
   LB:=0;
   UB:=Cmax;
   FindBound(LB,UB,T,C1,C2);
   i:=1;
   Cap1:=0;
   while C1.comp_a[i].end_time<=LB do
   begin
    Cap1:=Cap1+C1.comp_a[i].comp_pc;
    i:=i+1;
   end;
   j:=1;
   Cap2:=0;
   while C2.comp_a[j].end_time<=LB do j:=j+1;
   Ind:=j+1;
   while ind<=C2.nr_comp do
   begin
    Cap2:=Cap2+C2.comp_a[ind].comp_pc;
    ind:=ind+1;
   end;
   If C1.comp_a[i].comp_speed<>C2.comp_a[j].comp_speed then
   FindT:=(T.tasktime-Cap1-Cap2+C1.comp_a[i].comp_speed*c1.comp_a[i].beg_time-
           C2.comp_a[j].comp_speed*C2.comp_a[j].end_time)/
           (C1.comp_a[i].comp_speed-C2.comp_a[j].comp_speed)
   else FindT:=(UB-LB)/2;
 end;

 procedure AsRealProc(var sched:schedul_rst;proc_ind:ig;BegT,EnT:real;
                                                             T:task_rec0);
 {procedure assigns the task T to the real processor in the given interval}
 begin
  if (proc_ind<>0)then
  if (BegT<>EnT) then
  with sched.schedul[proc_ind]^ do
  begin
   nrts:=nrts+1;
   a[nrts].tnr:=T.nr;
   a[nrts].begint:=BegT;
   a[nrts].usestime:=EnT-BegT;
   a[nrts].endt:=pr_speed*(EnT-BegT);
  end;
 end;

 procedure assignTK(var sched:schedul_rst;Brd:Real;C:comp_proc;T:task_rec0);
 {procedure assigns the task T to the composite processor C in the
  interval [0,Brd]. Brd is a result of the FindT Function. As a result
  we assign the task T to the real procesors which parts of the processing
  time are composite parts of processor C}
 var
  i:Integer;
 begin
  i:=1;
  while (C.comp_a[i].end_time<Brd) do
  begin
   AsRealProc(sched,C.comp_a[i].pr_nr,C.comp_a[i].beg_time,
                                  C.comp_a[i].end_time,T);
   i:=i+1;
  end;
  if C.comp_a[i].end_time>=Brd then AsRealProc(sched,C.comp_a[i].pr_nr,
                                            C.comp_a[i].beg_time,Brd,T);
 end;

  procedure assignTK1(var sched:schedul_rst;Brd:real;C:comp_proc;T:task_rec0);
 {procedure assigns the task T to the composite processor C in the
  interval [Brd,Cmax]. As a result we assign the task T to the real procesors which
  parts of the processing time are composite parts of processor C}
 var
  i:Integer;
 begin
  i:=0;
  repeat
   i:=i+1;
  until (C.comp_a[i].end_time>=Brd)or (i=C.nr_comp);
  if (C.comp_a[i].end_time>Brd) then AsRealProc(sched,C.comp_a[i].pr_nr,Brd,
                                            C.comp_a[i].end_time,T);
  i:=i+1;
  while (i<C.nr_comp+1)and(C.comp_a[i].end_time<=Cmax) do
  begin
   AsRealProc(sched,C.comp_a[i].pr_nr,C.comp_a[i].beg_time,
                                  C.comp_a[i].end_time,T);
   i:=i+1;
  end;
 end;

 procedure AssignLast(var Tsk1:Task_rec0;var C:comp_proc);
 var
 i:Integer;
 X:Real;
 EndSched:Boolean;
 TskTime:Real;
 CompPc:Real;
 begin
  i:=0;
  Endsched:=False;
  TskTime:=Tsk1.tasktime ;
  repeat
   repeat
    i:=i+1;
   until C.comp_a[i].comp_pc>0;
   CompPc:=C.comp_a[i].comp_pc;
   X:= (tsktime+C.comp_a[i].beg_Time*C.comp_a[i].comp_speed)/C.comp_a[i].comp_speed;
   {if (tsktime<CompPc)or(tsktime=CompPc) then}
   begin
    AsRealProc(sched,C.comp_a[i].pr_nr,C.comp_a[i].beg_time,X,Tsk1);
    EndSched:=True;
   end; {else}
{   begin
    AsRealProc(sched,C.comp_a[i].pr_nr,C.comp_a[i].beg_time,C.comp_a[i].end_time,Tsk1);
    tsktime:=tsktime-CompPc;
    SetComp(C.comp_a[i],0,0,0,Cmax,0);
   end;                               }
  until EndSched;
 end;

 procedure CreateProc(Time:Real;var C1,C2:comp_proc;Flag:Boolean;T:task_rec0);
 {procedure creates two composite machines based on the machines C1,C2.
  The first one has speed s1 in the interval [time,Cmax] and s2 in the
  interval [0,time]. The capacity of this machine is S1+S2-T.tasktime.
  the second machine is created if processor c1 no.<nr_comp_proc.this
  machine has 0 speed}
 var
  i,j,k:Integer;
  CTab:array[1..max_comp]of comp_el;
  NewC:comp_el;
 begin
  For i:=1 to C2.nr_comp do
  begin
   if C2.comp_a[i].end_time>=time then break;
   CTab[i]:=C2.comp_a[i];
  end;
  if C2.comp_a[i].end_time>=time then
  begin
   CTab[i]:=C2.comp_a[i];
   CTab[i].end_time:=time;
   CTab[i].comp_pc:=(time-CTab[i].beg_time)*CTab[i].comp_speed;
  end;
  For j:=1 to C1.nr_comp do
  begin
   if C1.comp_a[j].end_time>=time then break;
  end;
  If C1.comp_a[j].end_time>time then
  begin
   i:=i+1;
   CTab[i]:=C1.comp_a[j];
   CTab[i].beg_time:=time;
   CTab[i].comp_pc:=(CTab[i].end_time-time)*CTab[i].comp_speed;
  end;
  if j<C1.nr_comp then
  for k:=j to C1.nr_comp do
  begin
   i:=i+1;
   CTab[i]:=C1.comp_a[j];
  end;
  C1.nr_comp:=i;
  C1.pc:=C1.pc+C2.pc-T.taskTime;
  for j:=1 to i do
  C1.comp_a[j]:=CTab[j];
  if Flag then
  begin
  { dispose(comprec.comp_pr[K1]);}
   dec(comprec.nr_comp_proc);
  end else
  SetDummy(C2);
 end;

 begin
   InitDuplik(TasksD,CompRec);
   with CompRec do
   with tasksd do
   with sched do
   begin
    gettime(hr,mn,sc,sc100);
    if not ExperInProgress then
    begin
     initsched(sched,1,nr_comp_proc);
     ExperInProgress:=True;
    end else
    resetsched(sched,1,nr_comp_proc);
    for i:=1 to nr_comp_proc do
    begin
     with comp_pr[i]^ do comp_a[i].comp_pc:=Cmax*comp_a[1].Comp_speed;
     sched.schedul[i]^.pr_speed:=comp_pr[i]^.comp_a[1].Comp_speed;
    end;
    SortProc(CompRec);
    SortTask(tasksd);
    Cmax:=FindCmax(tasksd,comprec);
    for i:=1 to nr_comp_proc do
    with comp_pr[i]^ do
    begin
     pc:=Cmax*comp_a[1].Comp_speed;
     comp_a[1].end_time:=Cmax;
     comp_a[1].comp_pc:=Cmax*comp_a[1].Comp_speed;
    end;
    repeat
     T:=task_v[1]^;
     i:=0;
     Dum:=False;
     repeat
      i:=i+1;
     until (comp_pr[i]^.pc<T.tasktime) or (i=nr_comp_proc);
     If (i=nr_comp_proc)and (comp_pr[i]^.pc>T.tasktime) then
      begin
       new(comp_pr[nr_comp_proc+1]);
       SetDummy(comp_pr[nr_comp_proc+1]^);
       K:=nr_comp_proc;
       inc(nr_comp_proc);
       K1:=nr_comp_proc;
       Dum:=True;
      end else
      begin
       K:=i-1;
       K1:=i
      end;
     DevT:=FindT(T,Comp_pr[K]^,Comp_pr[K+1]^);
     AssignTK(sched,DevT,Comp_pr[K]^,T);
     AssignTK1(sched,DevT,Comp_pr[K+1]^,T);
     CreateProc(DevT,Comp_pr[K]^,Comp_pr[K+1]^,Dum,T);
     SortProcC(comprec);
     DelTask(T);
    until (n=0)or((n=1)and OneNotDone(CompRec));
    if ((n=1)and OneNotDone(CompRec)) then
    begin
     T:=task_v[1]^;
     AssignLast(T,Comp_pr[1]^);
    end;
   end;
   workt:=getsec(hr,mn,sc,sc100);
   is_act_sch:=true;
   memory:=0;
   Sched.c:=Cmax;
  end;
end.






