{*****************************************************************************}
{*                                                                           *}
{*  Program : Simul                                                          *}
{*  Unit    : Hu                                                             *}
{*                                                                           *}
{*  This unit contains procedure implementing Hu scheduling algorithm.       *}
{*                                                                           *}
{*****************************************************************************}

Unit Hu;

{$C MOVEABLE DEMANDLOAD DISCARDABLE}

Interface

Uses Deklar,WinDos,WinPROCS,WinTypes,Tools;

Procedure Hu_(Var task: task_rec;Var comp: comp_rec; Var sched: schedul_rst;
                   Var workt: real;Var kp: ig);
{ Parameters given:
     task  - task set,
     comp - processor set,
     kp   - number of processors etc.,
  Parameters returned:
     sched   - schedule description,
     workt - how long schedule was computed}

Implementation

Type Tlevel = Array [0..MaxN] Of Byte;    {tablica poziomow zadan}
     Plevel = ^Tlevel;
     Tsucc = Array [1..MaxN] Of  fs;      {tablica zbiorow nastepnikow zadan}
     Psucc = ^Tsucc;
     Tno_exec = Array [1..MaxN] Of fs;    {tablica zbiorow nie wykonanych poprzednikow}
     Pno_exec = ^Tno_exec;

Procedure Hu_(Var task: task_rec;Var comp: comp_rec; Var sched: schedul_rst;
                   Var workt: real;Var kp: ig);

Var
   i,j,k: Byte;                       {identyfikatory elementow tablic}
   level: Plevel;
   hr,mn,sc,sc100: Word;              {czas}
   non_assigned: fs;              {zb. zad. nieprzydzielonych}
   execut: ^fs;
   succ: Psucc;
   no_exec: Pno_exec;
   wsk: ptrprem;
   t: Byte;                           {aktualny czas wykonywania zadan na obu maszynach}
   h: Byte;                           {ident. przydzielanych zadan}

Procedure set_levels(n: Byte; Var level: Plevel; Var succ: Psucc);
{This procedure sets levels of tasks}
Var
   lev,i,j: Byte;
   not_desc: fs;    {zb.zad., ktorym nie okreslono poziomu}
   desc: fs;        {zb.zad., ktorym okreslono poziom w danej iteracji}
Begin
  desc:=[];
  not_desc:=non_assigned;
  For i:=0 To n Do level^[i]:=0;
  lev:=1;
  While not_desc<>[] Do
  Begin
    For i:=1 To n Do
    Begin
      If ((level^[i]=0) And (succ^[i]=[])) Then
      Begin
        level^[i]:=lev;
        not_desc:=not_desc-[i];
        desc:=desc+[i];
      End;  {If}
    End; {For}
    For i:=1 To n Do
    If (i in desc) Then Begin
                          For j:=1 To n Do succ^[j]:=succ^[j]-[i];
                          desc:=desc-[i];
                        End;
    lev:=lev+1;
  End; {While}
End; {set_levels}

{-----------------------------}

Procedure najw_poz(Var x: Byte; Var level:Plevel; Var no_exec:Pno_exec);
{Procedura ta zwraca numer zadania, ktore ma najwyzszy}
{poziom, nalezy do zbioru nie przydzielonych zadan    }
{i ktorego wszystkie poprzedniki zostaly wykonane.    }
Var
   naj,q: Byte;
Begin
  naj:=0;
  For q:=1 To task.n Do
    If (q in non_assigned) And (level^[q]>level^[naj]) And (no_exec^[q]=[])
    And (Not(q in execut^))Then naj:=q;
  x:=naj;
End; {najw}
{-----------------------------}

Begin               {Hu}
  If HuConstrain<>outtree Then
  Begin
    memory:=0;
    New(level);
    New(succ);
    New(no_exec);
    New(execut);
    non_assigned:=[];                      {Tworzenie zbioru non_assigned i tablicy no_exec}
    For i:=1 To task.n Do
    Begin
      no_exec^[i]:=[];
      succ^[i]:=[];
    End;
    For i:=1 To task.n Do
    Begin
      non_assigned:=non_assigned+[i];
      If task.task_v[i]^.next<>nil Then
      Begin
        wsk:=task.task_v[i]^.next;
        While wsk<>nil Do
        Begin
          no_exec^[wsk^.nrprem]:=no_exec^[wsk^.nrprem]+[i]; {zad i jako poprzednik wsk^.nrprm}
          succ^[i]:=succ^[i]+[wsk^.nrprem];                 {wpis do tablicy nastepnikow}
          wsk:=wsk^.next;                                   {nastepny element a listy}
        End; {While}
      End; {If}
    End; {For}

    For i:=1 To task.n Do memory:=memory+SizeOf(no_exec^[i])+SizeOf(succ^[i]);
    memory:=memory+SizeOf(non_assigned)+SizeOf(execut^)+SizeOf(level^);

    t:=0;     {Ustawienie poczatkowych wartosci}
    With comp Do
      With sched Do
      Begin
        initsched(sched,1,nr_comp_proc);
        For i:=1 To nr_comp_proc Do schedul[i]^.nrts:=0;
      End; {with}
    i:=1;

    GetTime(hr,mn,sc,sc100);
    set_levels(task.n, level, succ);   {Okreslanie poziomow zadan}

    While non_assigned<>[] Do          {Przydzielenie zadan do procesorow}
    Begin
      With sched Do
      Begin
        execut^:=[];
        For j:=1 To comp.nr_comp_proc Do
        If non_assigned<>[] Then
        Begin
          najw_poz(h,level,no_exec);
          If h<>0 Then With schedul[j]^ Do {umieszczenie danych o zadaniu w tablicy sched dla procesora j}
          Begin
            a[i].tnr:=h;
            a[i].begint:=t;
            a[i].endt:=t+1;
            a[i].usestime:=1;
            end_time:=t+1;
            nrts:=nrts+1;
            non_assigned:=non_assigned-[h];
            execut^:=execut^+[h];
          End; {With}
        End; {If}
        For j:=1 To task.n Do
          If (j in execut^) Then For k:=1 To task.n Do no_exec^[k]:=no_exec^[k]-[j];
        t:=t+1;
        i:=i+1;
      End; {With}
    End; {While}
    sched.c:=t;
    sched.m:=comp.nr_comp_proc;
    workt:=GetSec(hr,mn,sc,sc100);
    Dispose(execut);
    Dispose(no_exec);
    Dispose(succ);
    Dispose(level);
    is_act_sch:=true;
  End
  Else
  Begin
    MessageBox(0,'Hu algorithm doesn`t solve'#10#13
                +'problems with outtree constrains.'#10#13#10#13
                +'No schedule achieved !','WARNING !',mb_OK+mb_IconExclamation);
    is_act_sch:=false;
  End;
  End;  {Hu}

End.



