{****************************************************************************}
{*                                                                          *}
{* Program : Simul                                                          *}
{* Unit    : K_D                                                            *}
{*                                                                          *}
{* This module implements an algorithm for preemptive scheduling of tasks   *}
{* requiring certain numbers of processors in parallel subject to deadlines *}
{* Processors are identical and optimality criterion is Lmax.               *}
{* Calls for the external program - a Linear Programming pocedure in file   *}
{* LP_SOLVE.EXE. (Run PIF file) The number of variables in the formulation  *}
{* of LP grows with m (the number of processors) exponentially              *}
{*                                                                          *}
{****************************************************************************}

unit K_D;

{$C MOVEABLE DEMANDLOAD DISCARDABLE}
{$X+}
interface

uses Deklar,Strings,Tools,ToolHelp,WinAPI,WinDos,WinPROCS,WinTypes,LPDane;

procedure Kd(var task:task_rec;var comp:comp_rec;var sched:schedul_rst;var workt:real;Var kp:ig);
{ Parametrs given:
     task - tasks set
     comp - processor set
     kp   - max. number of processors per multiprocessor task
  Parametrs returned:
     sched - schedule description
     workt - how long schedule was computed}

implementation

Procedure IOError(Var err: Word; Var des:String); Far; External 'ERRORS';
Procedure WinExecError(Var err: Word; Var des:String); Far; External 'ERRORS';

procedure Kd(var task:task_rec;var comp:comp_rec;var sched:schedul_rst;var workt:real;Var kp:ig);


          {This unit definitions}

{const MAXTASKNUM=100;    {max liczba zadan obu typow}
{      MAXINTERVAL=10;    {max liczba intervalow}
{      MAXPROCESSING=40;  {max dlugosc zadania}
{      R=10;              {biezaca gorna liczba intervalow}
{      STEP=10;           {max dlugosc intervalu}
{      PERTASK=10;        {ile eksperymentow na jedno zadanie}
{      Ttask=1;
{      Wtask=2;
{      out_file='lp.out';
{      in_file='lp.in';
{
type
 dane=record
       n,m,it    :integer;
       ni        :tab5;
       isens     :tab6 ;
       obj       :string[3];
       valg,vald :real8;
       c         :array[1..Mmax] of ^contr;
       v         :array[0..Nmax] of ^variable;
       blad      :integer;
      end;

 {coje=set of 1..255;}		 { przez to MAXTASK=255 }
 {wezptr=^wezel;
 wezel=record
        value1:fs;
        valuek:fs;
        next:wezptr
       end;
 string80=string[80];
 intptrptr=^intptr;
 intptr=record
          i:integer;
   	  ptr:intptrptr;
        end;

 TDd                =array [1..2,1..MAXTASKNUM] of real; {Duedate array }
 {Tivalend           =array[0..maxinterval] of real; {tablica koncow przedzilow}
 {TPt                =array [1..2,1..MAXTASKNUM] of real; {ProcessingTime array}
 {TStosy_Kombinacji  =array[1..MAXINTERVAL] of wezptr;
 {TLiczby_kombinacji =array[0..MAXINTERVAL] of integer;
 {Tstosy_przedzialow =array [1..2,1..MAXTASKNUM] of record
							i:integer;
							p:intptrptr;
                                                   end;
                  { nazwa moze byc mylaca, wskazania po zadaniach do list z # kombinacji zawierajacymi dane zadanie}

{var
    b:^dane;
    Xplain:boolean;
    s,o: string;
    str12: String[12];
    str8: String[8];
    str2: String[2];
    str1: Char;
    WinExErr,IOErr: Word;
    pos:  Array [0..127] Of Char;
    pos1: Array[0..127] Of Char;
    LP_Solv: Array [0..92] Of Char;
    cancelled: Boolean;
    headB,headA:wezptr;    {dwa stosy na kombinacje A-wszystko,B-tluste}
    {tse1,tsek  :fs;      {biezacy sklad kombinacji}
   { ma,n,k,n1,nk :integer; {ma maszyn, k procesorowosc,n1 zadan T,nk typu W}
    {curr_inval_begin:real; {begin time of current interval}
    {rl,tmp_nk,tmp_n1,      {lokalna wersja r}
     {		i,j,u,ui   :integer;	 {zm. robocze}
    {wynik:real;

    Dd:^TDd;
    ivalend:^Tivalend;
    Pt:^TPt;
    Stosy_Kombinacji:^TStosy_Kombinacji;
    Liczby_kombinacji:^TLiczby_kombinacji;
    Ile_kombinacji    :integer;{liczba wszystkich kombinacji we wszystkich przedzialach}
    {stosy_przedzialow:^Tstosy_przedzialow;
    koniec:real;


    num2num   : array[1..MAXTASKNUM] of integer;
    hr,mn,sc,sc100:word;
    file_name: Text;}


    {   ------Forward block- tego nie czytac to sie nie liczy-- }
function insystem(zadanie,tasktype:integer;now_interwal_start:real):boolean;forward;


procedure sort(l,r: integer);
var
 i,j: integer;
 y: pointer;
 x: real;
begin
 i:=l; j:=r; x:=task.task_v[(l+r) DIV 2]^.ddate;
 repeat
  while task.task_v[i]^.ddate<x do i:=i+1;
  while x<task.task_v[j]^.ddate do j:=j-1;
  if i<=j then
   begin
    y:=task.task_v[i];
    task.task_v[i]:=task.task_v[j];
    task.task_v[j]:=y;
    i:=i+1; j:=j-1;
   end;
 until i>j;
 if l<j then sort(l,j);
 if i<r then sort(i,r);
end;

procedure tta(nr_zad_tutaj,nr_paczki,typzadania:integer;var pierwszywolny:integer;koniec:real);
var i:integer;
    baza:integer;
begin
 if typzadania=Ttask then baza:=0 else baza:=n1;
 for i:=1 to task.task_v[num2num[baza+nr_zad_tutaj]]^.prcsnum do
  begin
   with sched.schedul[pierwszywolny]^ do
    begin
     end_time:=koniec+b^.v[nr_paczki]^.x;{end_time}
     inc(nrts);
     with a[nrts] do
      begin
       tnr:= num2num[baza+nr_zad_tutaj];
       begint:=koniec;
       endt:=begint+b^.v[nr_paczki]^.x;
       usestime:= b^.v[nr_paczki]^.x;
      end;
    end;
   inc(pierwszywolny);
  end;
end;




    {   ----------- OPERACJE NA STRUKTURACH DANYCH --------------}

procedure pop_tset(var tse1,tsek:fs;var head:wezptr);
var tmp:wezptr;
begin
 tse1:=head^.value1;
 tsek:=head^.valuek;
 tmp:=head^.next;
 dispose(head);
 head:=tmp;
end;

procedure push_tset(tse1,tsek:fs;var head:wezptr);
var tmp:wezptr;
    bx,by:byte;
begin
 new(tmp);
 tmp^.value1:=tse1;
 tmp^.valuek:=tsek;
 tmp^.next:=head;
 head:=tmp;
end;

function all_present(a:fs;tasktype,n:integer;curr_begin:real):boolean;
var i:integer;
    v:boolean;
begin
i:=0;
v:=true;
while (v and (i<n)) do
 begin
  i:=i+1;
  if (([i]*a=[]) and insystem(i,tasktype,curr_begin) )then v:=false;
 end;
all_present:=v;
end;



    { ------------ PROCEDURY GENERACJI KOMBINACJI -------------- }

procedure Ddgener(var intervals:integer;maxlen:integer);
var i,j:integer;
    events:fs;{zaznacza interwaly gdzie cos sie konczy}
    tmp_task:task_arr;
begin
 ivalend^[0]:=0;
 events:=[];
 intervals:=1;
{
-zadeklarowac tablice num2num[1..n] zapamietujaca numery w gornym module
 odpowiadajace numerom w tym module,
 numery uporzadkowane sa nastepujaco: pod indeksami 1..n1 sa nry zadan na
 gorze Ttaskow, pod nrami n1+1..n sa nry dla Wtask'ow -> zrobione}
{
-posortowac zadania 1...n wg task.task_v[1...n]^.ddate -> zrobione}

 tmp_task:=task.task_v;
 sort(1,n);


{-policzyc ile jest roznych linii krytycznych=intervals -> zrobine}
for i:=1 to task.n-1 do
  if task.task_v[i]^.ddate<>task.task_v[i+1]^.ddate then
   begin
    inc(intervals);
    ivalend^[intervals-1]:=task.task_v[i]^.ddate;
   end;
  ivalend^[intervals]:=task.task_v[task.n]^.ddate;

task.task_v:=tmp_task;
tmp_n1:=0;
tmp_nk:=0;
for i:=1 to n do if task.task_v[i]^.prcsnum>1 then
 begin
  inc(tmp_nk);
  Dd^[Wtask,tmp_nk]:=task.task_v[i]^.ddate;
  Pt^[Wtask,tmp_nk]:=task.task_v[i]^.tasktime  ;
  num2num[n1+tmp_nk]:=i;
 end
else
 begin
  inc(tmp_n1);
  Dd^[Ttask,tmp_n1]:=task.task_v[i]^.ddate    ;
  Pt^[Ttask,tmp_n1]:=task.task_v[i]^.tasktime;
  num2num[tmp_n1]:=i;
 end;
end;

function insystem(zadanie,tasktype:integer;now_interwal_start:real):boolean;
begin
 if Dd^[tasktype,zadanie]>now_interwal_start
  then insystem:=true
  else insystem:=false
end;

procedure nastepny1prosze(nowindex,wolne:integer;tse1,tsek:fs);
{nowindex-ostatnio sprawdzone zadanie,wolne-ile jeszcze procesorow wolnych,
tse1,tsek-aktualne zbiory dopuszczlne dla zad. 1- i k-procesorowych}
var i:integer;
    tsk:fs;
begin
if ((wolne>=1) and (insystem(nowindex,Ttask,curr_inval_begin))) then
 begin
  tse1:=tse1+[nowindex];
  push_tset(tse1,tsek,headA);
  for i:=nowindex+1 to n1 do nastepny1prosze(i,wolne-1,tse1,tsek);
 end
end;

procedure nastepnyKprosze(nowindex,wolne,krotnosc:integer;tse1,tsek:fs);
{nowindex-biezaco sprawdzane zadanie,wolne-ile jeszcze procesorow wolnych,
krotnosc - ile procesorow na zadanie,
tse1,tsek-aktualne zbiory dopuszczlne dla zad. 1- i k-procesorowych}
var i:integer;
    tsk:fs;
begin
if ((wolne>=krotnosc) and (insystem(nowindex,Wtask,curr_inval_begin))) then
 begin
  tsek:=tsek+[nowindex];
  push_tset(tse1,tsek,headA);
  for i:=1 to n1 do nastepny1prosze(i,wolne-krotnosc,tse1,tsek);
               (* generacja kombinacji 1-proc z biezaca kombinacja k-proc*)
  for i:=nowindex+1 to nk do nastepnyKprosze(i,wolne-krotnosc,krotnosc,tse1,tsek);
 end
end;



{procedura sito wybiera z wszystkich mozliwych tylko "najgestsze" kombinacje,
 do ktorych juz nic nie mozna dodac, bo albo jest zbyt malo wolnych procesorow
 albo wszyskie zadania sa w tej kombinacji. n1,nk- ile zad 1- i k-procesrowych,
 krotnosc-zadan wieloprocesorowych,wolne-liczba maszyn w systemie}

procedure sito(n1,nk,krotnosc,wolne:integer);
var tse1,tsek:fs;
    i,j:integer;
begin
while headA<>nil do
 begin
  pop_tset(tse1,tsek,headA);
  i:=wolne-cardinality(tse1)-krotnosc*cardinality(tsek);
  if (all_present(tsek,Wtask,nk,curr_inval_begin)
   or (i<krotnosc))
   and (all_present(tse1,Ttask,n1,curr_inval_begin) or (i=0))
    then push_tset(tse1,tsek,headB);
 end;
end;

function zmiennych_ile(inval:integer):integer;
var tmp:wezptr;
		i:integer;
begin
 tmp:=Stosy_kombinacji^[inval];
 i:=0;
 while not(tmp=nil) do
	begin
	 if (tmp^.value1<>[]) or (tmp^.valuek<>[])
		then inc(i);
	 tmp:=tmp^.next;
	end;
 zmiennych_ile:=i;
end; { Zmiennych_Ile }


procedure att(gdzie,co,tasktype:integer);
var tmp:intptrptr;
begin
 new(tmp);
 tmp^.i:=co;
 tmp^.ptr:=stosy_przedzialow^[tasktype,gdzie].p;
 stosy_przedzialow^[tasktype,gdzie].p:=tmp;
 inc(stosy_przedzialow^[tasktype,gdzie].i);
end; { att(ache) }



function valGLP:real;(* zwraca wynik GLP uwzgledniajac roznice miedzy valg i vald*)
begin (*zwraca gorsze - czyli wieksze*)
 if b^.valg>b^.vald then valGLP:=b^.valg else valGLP:=b^.vald;
end; (* valGLP *)




procedure setup_run;
var i,j,l,zm,blad,wolnyprocesor,
		ogr,nzero:integer;
		tmp:wezptr;
		tmpi:intptrptr;
    bx,by:byte;
    {heapark:pointer;}
    info:Integer;

{ pierwsza kombinacja w pierwszym przedziale = 1-wsza zmienna itd.
	pierwsza kombinacja w drugim=(ostatnia_w_1-wszym)+1 itd.}

begin
 ile_kombinacji:=0;
 liczby_kombinacji^[0]:=0;
 for i:=1 to rl do
  begin
   liczby_kombinacji^[i]:=zmiennych_ile(i);
   ile_kombinacji:=ile_kombinacji+liczby_kombinacji^[i];
	end;
 for i:=1 to nk do
	begin
	 stosy_przedzialow^[Wtask,i].p:=nil;
	 stosy_przedzialow^[Wtask,i].i:=0;
	end;
 for i:=1 to n1 do
	begin
	 stosy_przedzialow^[Ttask,i].i:=0;
	 stosy_przedzialow^[Ttask,i].p:=nil;
	end;
 zm:=0;
 for i:=1 to rl do 		{ przekstalcenie przedzialy-kombinacje -> }
	begin								{ zadania-kombinacje }
	 tmp:=stosy_kombinacji^[i];
	 zm:=zm+liczby_kombinacji^[i-1];
	 for j:=1 to liczby_kombinacji^[i] do
		begin
		 for l:=1 to n1 do if tmp^.value1*[l]<>[] then att(l,zm+j,Ttask);
		 for l:=1 to nk do if tmp^.valuek*[l]<>[] then att(l,zm+j,Wtask);
		 tmp:=tmp^.next;
		end;
	end;

 b^.m:=rl+n1+nk; 		{ograniczen tyle ile zadan i przedzialow}
 b^.n:=ile_kombinacji+1;	{zmiennych tyle ile kombinacji +Lmax}

 for i:=0 to b^.n do			{ f.celu }
	begin
	 getmem(b^.v[i],sizeof(variable));
	 with b^.v[i]^ do
		begin
		 xinf:=0;
		 x:=0;
		 xsup:=borne;
		 g:=0
		end;
  end;
 b^.v[b^.n]^.g:=1;	{Lmax}
 b^.obj:='min';
												{ wszystkie zmienne mieszcza sie}
												{ w poszczegolnych przedzialach }

 b^.ni[1]:=liczby_kombinacji^[1]+1;
 Getmem(b^.c[1],4*sizeof(real8)+b^.ni[1]*sizeof(tab));
 for j:=1 to b^.ni[1]-1 do
	begin
	 b^.c[1]^.t[j].col:=j;
	 b^.c[1]^.t[j].ele:=1;
	end;
 b^.c[1]^.t[b^.ni[1]].col:=b^.n;
 b^.c[1]^.t[b^.ni[1]].ele:=-1;
 b^.isens[1]:={0}-1;
 b^.c[1]^.b:=ivalend^[1];		{I-wszy przedzial eXtra- ma Lmax'a }

 zm:=0;
 for i:=2 to rl do
	begin
	 b^.ni[i]:=liczby_kombinacji^[i];
	 zm:=zm+liczby_kombinacji^[i-1];
	 Getmem(b^.c[i],4*sizeof(real8)+b^.ni[i]*sizeof(tab));
	 for j:=1 to b^.ni[i] do
		begin
		 b^.c[i]^.t[j].col:=zm+j;
		 b^.c[i]^.t[j].ele:=1;
		end;
	 b^.isens[i]:={0}-1;
	 b^.c[i]^.b:=ivalend^[i]-ivalend^[i-1];
	end;
												{wszystkie zadania sa wykonane}
												{ Ttasks }
 for i:=rl+1 to rl+n1 do
	begin
	 b^.ni[i]:=stosy_przedzialow^[Ttask,i-rl].i;
	 Getmem(b^.c[i],4*sizeof(real8)+b^.ni[i]*sizeof(tab));
	 tmpi:=stosy_przedzialow^[Ttask,i-rl].p;
	 for j:=1 to b^.ni[i] do
		begin
		 b^.c[i]^.t[j].col:=tmpi^.i;{kolumna}
		 b^.c[i]^.t[j].ele:=1;
		 tmpi:=tmpi^.ptr;
		end;
	 b^.isens[i]:={0}1;
	 b^.c[i]^.b:=Pt^[Ttask,i-rl];
	end;
												{ Wtasks }

 for i:=rl+n1+1 to rl+n1+nk do
	begin
	 b^.ni[i]:=stosy_przedzialow^[Wtask,i-rl-n1].i;
	 Getmem(b^.c[i],4*sizeof(real8)+b^.ni[i]*sizeof(tab));
	 tmpi:=stosy_przedzialow^[Wtask,i-rl-n1].p;
	 for j:=1 to b^.ni[i] do
		begin
		 b^.c[i]^.t[j].col:=tmpi^.i;{kolumna}
		 b^.c[i]^.t[j].ele:=1;
		 tmpi:=tmpi^.ptr;
		end;
	 b^.isens[i]:=1;
	 b^.c[i]^.b:=Pt^[Wtask,i-rl-n1];
	end;

 Xplain:=true;


  {********** Execute External program - LP package ***********}

    b^.blad:=0; {=blad}
{$I+}
    {Zapis danych do pliku wejsciowego (in_file) dla zewn. procedury LP}
    Assign(file_name,WorkDir+'\'+in_file);
{$I-}
    Rewrite(file_name);
{$I+}
    IOErr:=IOResult;
    If IOErr<>0 Then
     Begin
       Str(IOErr,o);
       o:='IO Error ' + o + ' Occured While'#10#13
         +'Writing Data for External Linear Programing Procedure';
       StrPCopy(pos,s);
       MessageBox(0,PChar(@pos),'IO Error',mb_OK+mb_IconExclamation);
     End  {If IOErr<>0}
    Else
     Begin
       With b^ Do
       Begin
         Write(file_name, obj:3);
         Writeln(file_name, ':  Lmax;');   {1- linia - f.celu}
         For i:=1 To (ni[1]-2) Do
             Writeln(file_name, 'x' , c[1]^.t[i].col,' + ');
         Writeln(file_name, 'x' , c[1]^.t[ni[1]-1].col, ' - ');
         Writeln(file_name, 'Lmax');
         Write(file_name, ' <= ');
         Writeln(file_name, c[1]^.b:8 , ';');{2 - linia ograniczenie z Lmax}
         For j:=2 To m Do {dla zwyklych ograniczen}
         Begin
           For i:=1 To (ni[j]-1) Do
               Writeln(file_name, 'x' ,c[j]^.t[i].col,' + ');
           If ni[j] > 1 Then Writeln(file_name, 'x' ,c[j]^.t[i+1].col);
           If ni[j] = 1 Then Writeln(file_name, 'x' ,c[j]^.t[1].col);
           If isens[j] = 1 Then Write(file_name, ' >= ');
           If isens[j] = -1 Then Write(file_name, ' <= ');
           Writeln(file_name, c[j]^.b:8 , ';');    {wspolczynnik b}
         End; {For j}
       End; {With b^}
       Close(file_name);
     End; {Else IF IOErr<>0} {przesunac na koniec procedury obslugi LP]

{--- koniec operacji plikowych - gotowy do wywolania solvera ---}


    {Wywolanie zewnetrznej procedury LP}
    cancelled:=true;
    Repeat
    StrPCopy(LP_Solv,LPPath+'\'+LPName);
    WinExErr:=WinExec(PChar(@LP_Solv),sw_ShowNormal);
    info:= MessageBox(0,'Calling External Linear Programing Procedure'#10#13'Strange '
                         +'characters may appear','Information',mb_OK+mb_IconInformation+mb_ApplModal);
    If WinExErr<32 Then failure := true;
    Until cancelled;


    {Odczyt danych (rozwiazania) z pliku wejsciowego (out_file)}

    Assign(file_name,WorkDir+'\'+out_file);
{$I-}
    Reset(file_name);
{SI+}
    failure := false;
    If IOResult <>0 Then failure := true
    Else
    Begin
      Read(file_name, str5);
      If str5 <> 'Value' Then failure := true
    End;
    If (failure = true )Then MessageBox(0,'IO Error occured while reading results'#10#13
                       +'from External Linear Programing Procedure',
                       'IO Error',mb_OK+mb_IconExclamation)
    Else
    Begin
      Readln(file_name, str23, b^.valg); {wartosc f. celu pierwotnej}
      Readln(file_name, str5,  b^.v[b^.n]^.x); {wartosc zmiennej Lmax}
      b^.vald := b^.valg;
      For i:=1 To b^.n-1 Do
      Begin
        Readln(file_name, str1, index, value);
        b^.v[index]^.x := value;
      End;
      b^.blad := 0;
      Close(file_name);


  {************************************************************}

      blad:= doserror+100*b^.blad;
      str(blad:5,s);
      StrPCopy(pos1,s);
      StrPCopy(pos,'GRADIENT ERROR: ');
      StrCat(pos,pos1);
      StrCat(pos,#10#13);

      if blad<>0 Then
        Begin
          str(workt:6:4,o);
          StrPCopy(pos1,o);
          StrCat(pos,'TIME= ');
          StrCat(pos,pos1);
          StrCat(pos,#10#13);
          StrPCopy(pos1,'No schedule achieved !');
          StrCat(pos,pos1);
          MessageBox(0,pos,'Critical Error',mb_OK+mb_IconExclamation);
        End {if blad<>0}
        Else
        Begin
          wynik:=valglp;

          zm:=0;
          koniec:=0;

          for i:=1 to rl do 		{ po przedzialach }
          begin								{ }
            tmp:=stosy_kombinacji^[i];
            zm:=zm+liczby_kombinacji^[i-1];
            for j:=1 to liczby_kombinacji^[i] do
            begin
              if b^.v[zm+j]^.x>0 then
   		  begin
                    wolnyprocesor:=1;
                    for l:=1 to n1 do if tmp^.value1*[l]<>[] then tta(l,zm+j,Ttask,wolnyprocesor,koniec);
	            for l:=1 to nk do if tmp^.valuek*[l]<>[] then tta(l,zm+j,Wtask,wolnyprocesor,koniec);
                    koniec:=koniec+b^.v[zm+j]^.x;
		  end; {if b^.v[zm+j]^.x>0}
              tmp:=tmp^.next;
            end; {for j}
          end;  {for i}
          is_lmax_crit:=true;
          is_act_sch:=true;
        End; {Else if blad<>0}
    End; {Else if IOResult<>0}

      for i:=0 to ile_kombinacji+1 do freemem(b^.v[i],sizeof(variable));
      for i:=1 to rl+n1+nk do freemem(b^.c[i],4*sizeof(real8)+b^.ni[i]*sizeof(tab));

      for i:=1 to nk do
      while stosy_przedzialow^[Wtask,i].p<>nil do
      begin
        tmpi:=stosy_przedzialow^[Wtask,i].p;
        stosy_przedzialow^[Wtask,i].p:=stosy_przedzialow^[Wtask,i].p^.ptr;
        dispose(tmpi);
      end;
      for i:=1 to n1 do
      while stosy_przedzialow^[Ttask,i].p<>nil do
      begin
        tmpi:=stosy_przedzialow^[Ttask,i].p;
        stosy_przedzialow^[Ttask,i].p:=stosy_przedzialow^[Ttask,i].p^.ptr;
        dispose(tmpi);
      end;
end;	{ SetUp_Run }


begin               {Schedule}
 memory:=0;
 New(b);
 New(Dd);
 New(ivalend);
 New(Pt);
 New(Stosy_Kombinacji);
 New(Liczby_kombinacji);
 New(stosy_przedzialow);

 ma:=comp.nr_comp_proc;(*liczba zadan*)
 k:=kp;                (*iloprocesorowe zadania*)
 n:=task.n;            (*liczba zadan*)
 nk:=0;
 for i:=1 to n do if task.task_v[i]^.prcsnum>1 then inc(nk);
 n1:=n-nk;             (*n1 liczba zadan jednoprocesorowych*)
 rl:=r;
 for i:=1 to MAXINTERVAL do stosy_kombinacji^[i]:=nil;
 gettime(hr,mn,sc,sc100);
 Ddgener(rl,step);
 initsched(sched,1,comp.nr_comp_proc);
 for j:=1 to rl do
  begin
   headA:=nil;
   headB:=nil;
   tse1:=[];
   tsek:=[];
   curr_inval_begin:=ivalend^[j-1];
   for i:=1 to n1 do nastepny1prosze(i,ma,[],[]);
   (*generacja wszystkich zbiorow dopuszczalnych zadan 1-procesorowych
   przy pustym zbiorze zadan k-procesorowych*)
   (*if readkey='?' then stackstate(true,headA);*)
   for i:=1 to nk do nastepnyKprosze(i,ma,k,[],[]);
   (* generacja przy niepustym zbiorze zad. k-procesorowych *)
   (*if readkey='?' then stackstate(true,headA);*)
   sito(n1,nk,k,ma);         {odchudzenie zbioru kombinacji}
   (*stackstate(false,headB,j);*)
   (*stackstate(true,headB,j);*)
   Stosy_kombinacji^[j]:=headB;
  end;


 is_lmax_crit:=false;
 is_act_sch:=false;

 Setup_Run;
 workt:=getsec(hr,mn,sc,sc100);
 task.nr_k:=nk;
 sched.c:=koniec;
 sched.lmax:=wynik;
 for i:=1 to MAXINTERVAL do
 while stosy_kombinacji^[i]<>nil do
  begin
   headA:=stosy_kombinacji^[i];
   stosy_kombinacji^[i]:=stosy_kombinacji^[i]^.next;
   dispose(headA);
  end;
 memory:=memory+SizeOf(b^)+Sizeof(ivalend^)+SizeOf(Stosy_Kombinacji);
 memory:=memory+SizeOf(Liczby_Kombinacji)+SizeOf(Stosy_przedzialow);
 Dispose(b);
 Dispose(Dd);
 Dispose(ivalend);
 Dispose(Pt);
 Dispose(Stosy_Kombinacji);
 Dispose(Liczby_kombinacji);
 Dispose(stosy_przedzialow);

end;

end.








