{*****************************************************************************}
{*                                                                           *}
{*  Program : Simul                                                          *}
{*  Unit    : MCof                                                           *}
{*                                                                           *}
{*  This module implements Muntz-Coffman's algorithm for preemptive          *}
{*  scheduling of tasks with precedence constraints graph of the forest      *}
{*  form on any number of processors or any precedence constraints graph     *}
{*  on two processors. Processors are identical and optimization criterion   *}
{*  is Cmax.                                                                 *}
{*                                                                           *}
{*****************************************************************************}

{$C MOVEABLE DEMANDLOAD DISCARDABLE}

unit MCof;

interface

uses WinDos,Deklar,Tools;

procedure Mcofs(task:task_rec; comp:comp_rec; var sched:schedul_rst;var workt:real;kp:ig);
{ Parametrs given:
     task - tasks set,
     comp - processor set,
     kp   - number of processors executing widest multiprocessor task,
  Parametrs returned:
     sched - schedule description,
     workt - how long schedule was computed}


implementation

procedure Mcofs(task:task_rec; comp:comp_rec; var sched:schedul_rst;var workt:real; kp:ig);
var
   level : array[1..MaxN] of ig; { Tasks' levels array }
   ptime : array[1..MaxN] of ig; { Time to the end of task execution array }
   rdy : array[1..MaxN] of ig;   { Array of ready for execution tasks }
   nrdy:ig;                      { Number of tasks ready for execution }
   max,i,j,k,l:ig;
   ch:char;
   knc:boolean;
   hr,mn,sc,sc100:word;
   time:real;                     { current time of assigning tasks to processors }
   hl:real;                       { How long is task to be executed? }
   newlevels:boolean;


procedure int_task_time(var t:task_rec);
{ This procedure removes fractional parts of processing times
(assumption of Muntz-Coffman's algorithm is processing times are integers)}
var
	i:ig;
begin
 with t do for i:=1 to n do
 task_v[i]^.tasktime:=int(task_v[i]^.tasktime);
end;






procedure which_rdy(task:task_rec);
{ This procedure finds tasks ready for processing
from among not executed tasks}

var
  rdyt : array[1..MaxN] of boolean;
  i:ig;
  nx:ptrprem;

begin
 with task do
  begin
   for i:=1 to n do rdyt[i]:=TRUE;
   for i:=1 to n do
   if ptime[i]=0 then rdyt[i]:=FALSE
   else with task_v[i]^ do
    begin
     nx:=next;
     while nx<>nil do
      begin
       rdyt[nx^.nrprem]:=FALSE;
       nx:=nx^.next;
      end;
    end;
   nrdy:=0;
   for i:=1 to n do if rdyt[i]=TRUE then
    begin
     nrdy:=nrdy+1;
     rdy[nrdy]:=i;
    end;
  end;
end;




procedure what_levels(task:task_rec);
{ This procedure finds levels of all tasks.}
var
   i:integer;

function what_l(t:ig):ig;
{ This function finds level of the task t}
var
   i,sc,max,l:ig;
   nx:ptrprem;
begin
 with task do
  begin
   if level[t]=0 then
    begin
     nx:=task_v[t]^.next;
     sc:=0;
     while nx<>nil do
      begin
       sc:=sc+1;
       nx:=nx^.next;
      end;                 { sc - number of direct successors }
     if sc=0 then what_l:=trunc(task_v[t]^.tasktime)
     else
      begin
       nx:=task_v[t]^.next;
       max:=0;
       for i:=1 to sc do
        begin
         l:=what_l(nx^.nrprem);
         if l>max then max:=l;
         nx:=nx^.next;
        end;             { max - the highest level of any successor }
       what_l:=trunc(task_v[t]^.tasktime)+max;
      end;
    end
   else what_l:=level[t];
  end;
end;

begin      { what_levels }
 with task do
  begin
   for i:=1 to n do level[i]:=0;    { 0 - level not calculated }
   for i:=1 to n do level[i]:=what_l(i);
  end;
end;

procedure fill_tasks(var sched:schedul_rst; tm:real; nut:ig);
{ This procedure assigns nut ready tasks to processors
  in time interval tm long}
var
   np,i:ig;
   tt:real;
   equal,low,great:boolean;   { Variables for McNaughton's (sub-)algorithm }
   cmax:real;

begin
 with sched do
  begin
   cmax:=tm;
   if nut<=m then  { Give every task a unit of processing time }
		begin
	   for i:=1 to m do
     if (i<=nut) then
      begin
       with schedul[i]^ do
  	    begin
   	     nrts:=nrts+1;
      	 if (nrts>1) or (time<>0) then a[nrts].begint:=time
         else a[nrts].begint:=0;
	       with a[nrts] do
   	      begin
           endt:=begint+tm;
   	       usestime:=tm;
           tnr:=rdy[i];
           end_time:=endt;
 	        end;
   	    end
      end
    end
   else  { Apply McNaughton's algorithm }
    begin
     cmax:=tm*nut/m;
     np:=1;
     for i:=1 to nut do
      begin
       with schedul[np]^ do
        begin
         nrts:=nrts+1;
         if nrts>1 then
          if a[nrts-1].endt>time then a[nrts].begint:=a[nrts-1].endt
          else a[nrts].begint:=time
         else a[nrts].begint:=0;
         with a[nrts] do
          begin
           tnr:=rdy[i];
           equal:=(begint+tm)=(time+cmax);   { A task fits in the interval 0..cmax }
           low:=(begint+tm)<(time+cmax);     { The tasks fills completely 0..cmax interval }
           great:=(begint+tm)>(time+cmax);   { The (end of) task doesn't fit in 0..Cmax interval}
           if low then
            begin
             endt:=begint+tm;
             usestime:=tm;
            end;
           if equal then
            begin
             endt:=time+cmax;
             usestime:=endt-begint;
             end_time:=endt;
             np:=np+1;
            end;
          end;
        end;
       if great then
        begin
         with schedul[np]^ do
         with a[nrts] do
          begin
           endt:=time+cmax;
           usestime:=endt-begint;
           end_time:=endt;
           tt:=tm-usestime;   { Time till the end of execution of the task}
          end;
         np:=np+1;
         with schedul[np]^ do
          begin
           nrts:=nrts+1;
           with a[nrts] do
            begin
             begint:=time;
             endt:=begint+tt;
             usestime:=tt;
             tnr:=rdy[i];
            end;
          end;
        end;
      end;
    end;
  end;
 time:=time+cmax;
end;


begin         { Schedulding }
 time:=0;
 int_task_time(task);
 with comp do
 with sched do
 with task do
  begin
	{if is_act_sch then
         release(heapptr);
      mark(heapptr);}
   gettime(hr,mn,sc,sc100);
   m:=nr_comp_proc;
   initsched(sched,1,nr_comp_proc);
   what_levels(task);
   time:=0;
   for i:=1 to n do ptime[i]:=trunc(task_v[i]^.tasktime);
   which_rdy(task);
   while nrdy>0 do
    begin

 { Sorting available tasks by level }
     repeat
      for i:=1 to nrdy-1 do
      for j:=i+1 to nrdy do
      if level[rdy[i]]<level[rdy[j]] then
       begin
        k:=rdy[i];
        rdy[i]:=rdy[j];
        rdy[j]:=k;
       end;

 { Finding number of the highest level tasks }
      max:=level[rdy[1]];
      l:=1;
      knc:=FALSE;
      repeat
       if (rdy[l+1]<=n) and (l<nrdy) then
        if max=level[rdy[l+1]] then l:=l+1
        else knc:=TRUE
       else knc:=TRUE;
      until knc;
      while (l<m) and (l<nrdy) do
       begin
        l:=l+1;
        max:=level[rdy[l]];
        knc:=FALSE;
        repeat
         if (rdy[l+1]<=n) and (l<nrdy) then
          if max=level[rdy[l+1]] then l:=l+1
          else knc:=TRUE
         else knc:=TRUE;
        until knc;
       end;
      if l>nrdy then l:=nrdy;
      if (rdy[l+1]<=n) and (l<nrdy) then hl:=level[rdy[l]]-level[rdy[l+1]]
      else hl:=maxint;

 { Finding minimal task execution time }
      for i:=1 to l do if ptime[rdy[i]]<hl then hl:=ptime[rdy[i]];
      fill_tasks(sched,hl,l);
      newlevels:=FALSE;
      for i:=1 to l do
       begin
        level[rdy[i]]:=level[rdy[i]]-trunc(hl);
        ptime[rdy[i]]:=ptime[rdy[i]]-trunc(hl);
        if ptime[rdy[i]]=0 then newlevels:=TRUE;
       end
     until newlevels;
     which_rdy(task);
    end;
   c:=0;
   for i:=1 to m do if c<schedul[i]^.end_time then c:=schedul[i]^.end_time;
   for i:=1 to m do schedul[i]^.end_time:=c;
   workt:=getsec(hr,mn,sc,sc100);
   memory:=0;
  end;
end;


end.