{*****************************************************************************}
{*                                                                           *}
{*  Program : Simul                                                          *}
{*  Unit    : MKProc                                                         *}
{*                                                                           *}
{*  This unit implements algorithm scheduling multiprocessor tasks ie.       *}
{*  requireing more than one processor at the time for its processing.       *}
{*  The algorithm handles case of 1-,...,i-,j-,...k- processor tasks         *}
{*  on condition that j/iZ+ (is positive integer). Processor set consists   *}
{*  of m uniform machines. Optimality criterion is Cmax.                     *}
{*                                                                           *}
{*****************************************************************************}

{$I PCoption}
{$C MOVEABLE DEMANDLOAD DISCARDABLE}
unit mkproc;

interface

uses WinDos,Deklar,Tools;

procedure Mkprocs(var tasks:task_rec;var comp:comp_rec;var sched:schedul_rst;var workt:real;kp:ig);
{ Parametrs given:
     task - tasks set
     comp - processor set
     kp -mx. number of processors per multiprocessor task
  Parametrs returned:
     sched - schedule description
     workt - how long schedule was computed}

implementation

procedure Mkprocs(var tasks:task_rec;var comp:comp_rec;var sched:schedul_rst;var workt:real;kp:ig);

const
	max_k_group=30;
	maxdelt=50;

type
	nrdp_arr=array[1..MaxM]of ig;
	dp_arr=array[1..MaxM]of real;
	head_list_ptr=^head_list;
	head_list=record
  		   first_proc:ig;
		   last_proc:ig;
		   next_h_l:head_list_ptr;
		  end;

var
	cmax:real;   									    { current schedule length }
	k_arr:array[1..max_k_group] of ig;{ This array carries numbers of necessary }
                                    { processors in appropriate tasks groups}
	nr_k_arr:ig;									    { Number of k-processor groups }
	nr_t_gr:array[0..max_k_group] of ig;   { Number of tasks in the group}
	hr,mn,sc,sc100:word;
	tmp_comp:comp_procs;					    { Temporary copy of processor set }
	tmp_proc:comp_proc;
        t_ra:task_rec1;
	tmp_task:task_rec0;
	k_task:task_arr;                  { K-processor tasks }
	m_comp:ig;								        { Number of processors }
	tmp_k:ig;
	stake:head_list_ptr;
	x,i,j,l:ig;
	tab_nrdp:nrdp_arr;		{ Indices of processors with dead processing capacity }
	tab_dp:dp_arr;				{ Dead processing capacities array }
	nrpd:ig;						  { Number of dead processing capacity cases}
	a_cmax:array[1..maxdelt] of real;{ Array recording computed versions of Cmax }
	nr_cmax:ig;											 { Number of versions of Cmax (iterations }
                                   { of the algorithm) }
	proc_arr:array[1..MaxM] of head_list_ptr;	{ Array of pointers to processor }
                                  { indices sub-lists (real processors }
                                  { contributing in given composite processor)}
	nr_list:ig;											{ Number of sub-list }


procedure chack_tasks(tasks:task_rec);
{ This procedure checks corectness of data and fills k_arr array}
var
	i,j,k:ig;
	knc:boolean;
begin
 for i:=1 to max_k_group do
	begin
	 k_arr[i]:=0;
	 nr_t_gr[i]:=0;
	end;
 nr_k_arr:=0;
 with tasks do
	begin
	 for i:=1 to n do
	 with task_v[i]^ do
		begin
		 j:=1;
		 knc:=false;
		 while (not knc) and (j<=nr_k_arr) do
			begin
			 if k_arr[j]=prcsnum then
				begin
				 knc:=true;
				 inc(nr_t_gr[j]);
				end
			 else inc(j);
			end;
		 if not knc then
			begin
			 inc(nr_k_arr);
			 k_arr[nr_k_arr]:=prcsnum;
			 inc(nr_t_gr[nr_k_arr]);
			end;
		end;
	 for i:=1 to nr_k_arr-1 do				{ Sorting }
	 for j:=i+1 to nr_k_arr do
	 if k_arr[i]>k_arr[j] then
	  begin
		 k:=k_arr[i]; k_arr[i]:=k_arr[j]; k_arr[j]:=k;
		 k:=nr_t_gr[i]; nr_t_gr[i]:=nr_t_gr[j]; nr_t_gr[j]:=k;
		end;
	 k:=k_arr[2] div k_arr[1]; 				{ Checking tasks }
	end;
end;

procedure init_list(n_pr:ig);
{ This procedure initializes list structure ( n_pr - number of processors)}
var
 i:ig;
begin
 nr_list:=1;
 new(proc_arr[1]);
 Inc(memory,SizeOf(head_list)+SizeOf(proc_arr[1]));
 with proc_arr[1]^ do
   begin
    next_h_l:=nil;
    first_proc:=1;
    last_proc:=n_pr;
   end;
end;

procedure split_list(nr_e:ig);
{ This procedure divides processor indices list into appropriate sub-lists}
{ nr_e - number of processors necessary for currently analysed tasks type }
var
	i,j,k:ig;
	nr_f:ig;		{ Number of processors required by previously analysed type of tasks }
	baza:ig;
	wsk,wskh,wskp:head_list_ptr;
begin
 nr_f:=proc_arr[1]^.last_proc;
 k:=nr_f div nr_e;
 for i:=nr_list downto 2 do
	begin
	 proc_arr[(i-1)*k+1]:=proc_arr[i];
	 proc_arr[i]:=nil;
	end;
 for i:=1 to	nr_list do
	begin
	 for j:=1 to k do
		begin
		 wskh:=proc_arr[(i-1)*k+1];
		 wsk:=proc_arr[(i-1)*k+j];
		 if j>1 then wskp:=proc_arr[(i-1)*k+j-1];
		 while wskh<>nil do
    		  begin
		   if j=1 then
		    begin
		     wsk^.last_proc:=wsk^.first_proc+nr_e-1;
		     wsk:=wsk^.next_h_l;
		    end
		   else
		    begin
		     if wskh=proc_arr[(i-1)*k+1] then
		      begin
		       new(proc_arr[(i-1)*k+j]);
                       Inc(memory,SizeOf(head_list)+SizeOf(proc_arr[(i-1)*k+j]));
		       wsk:=proc_arr[(i-1)*k+j];
		      end
		     else
		      begin
		       new(wsk^.next_h_l);
                       Inc(memory,SizeOf(wsk^.next_h_l));
		       wsk:=wsk^.next_h_l;
		      end;
		   wsk^.first_proc:=wskp^.last_proc+1;
		   wsk^.last_proc:=wskp^.last_proc+nr_e;
		   wsk^.next_h_l:=nil;
		   wskp:=wskp^.next_h_l;
		  end;
		 wskh:=wskh^.next_h_l;
	        end;
	    end;
	end;
	nr_list:=nr_list*k;
end;


procedure join_list(nr_l:ig);
{ This procedure joins two neighbouring lists of processor indices}
{ into one list of indices }
var
	i,j:ig;
	wsk:head_list_ptr;
begin
 if nr_l<nr_list then
	begin
	 wsk:=proc_arr[nr_l];
	 while wsk^.next_h_l<>nil do wsk:=wsk^.next_h_l;
	 wsk^.next_h_l:=proc_arr[nr_l+1];
	 for i:=nr_l+1 to nr_list-1 do proc_arr[i]:=proc_arr[i+1];
	 proc_arr[nr_list]:=nil;
	 nr_list:=nr_list-1;
	end;
end;



function tsk(tasks:task_rec; nr:ig; k:ig):ig;
{ This procedure finds nr-th real processor in k-processor indices data structure}
var
	i,j:ig;
begin
 with tasks do
	begin
	 j:=1;
	 for i:=1 to nr do
		begin
		 while task_v[j]^.prcsnum<>k do inc(j);
		 inc(j);
		end;
	 tsk:=j-1;
	end;
end;

procedure setrst(var sched:schedul_rst;var comp:comp_rec);
var i:ig;
begin
 with sched do
 for i:=1 to m do
  begin
   with comp.comp_pr[i]^ do
	 with schedul[proc_nr]^ do
    begin
     end_time:=0;
     nrts:=0;
     pr_speed:=speed;
    end;
   c:=0;
  end;
end;

procedure cap_end_time(cmax:real;n:ig;var comp_p:comp_procs);
var i:ig;
begin
 for i:=1 to n do
 with comp_p[i]^ do
 with comp_a[1] do
  begin
   pc:=cmax*speed;
   beg_time:=0;
   end_time:=cmax;
   comp_pc:=pc;
  end;
end;

function calculate_C(tasks:task_rec;comp:comp_rec):real;
{ This function returns C as a maximum over C(1),..., C(k)}
var
	c,sum1,sum2,sum:real;
	i,j,l,k,g,ix:ig;
	tsk0:task_rec0;
	n_task,m_proc:ig;
	tmp_task:task_arr;
	heapPTR:pointer;
begin
 {mark(heapPtr);}
 with tasks do
 with comp do
	begin
	 c:=0;
	 for l:=1 to nr_k_arr do
		begin
		 m_proc:=nr_comp_proc div k_arr[l];			{ m_proc - number of processors}
		 n_task:=0;
		 for i:=l to nr_k_arr do
                  n_task:=n_task+nr_t_gr[i]*(k_arr[i] div k_arr[l]); {n_task-# of tasks}
		 for i:=1 to n_task do
                  begin
                   new(tmp_task[i]);
                   Inc(memory,SizeOf(task_rec0))
                  end;
		 ix:=1;
		 for i:=1 to nr_t_gr[l] do
			begin
			 tmp_task[ix]^:=task_v[tsk(tasks,i,k_arr[l])]^;
			 ix:=ix+1;
			end;
		 for i:=l+1 to nr_k_arr do
		 for j:=1 to nr_t_gr[i] do
     for k:=1 to k_arr[i] div k_arr[l] do
			begin
			 tmp_task[ix]^:=task_v[tsk(tasks,j,k_arr[i])]^;
			 ix:=ix+1;
			end;
		 for i:=1 to n_task-1 do					{ Sorting tasks by processing time }
		 for j:=i+1 to n_task do
     if tmp_task[i]^.tasktime<tmp_task[j]^.tasktime then
		  begin
			 tsk0:=tmp_task[i]^;
			 tmp_task[i]^:=tmp_task[j]^;
			 tmp_task[j]^:=tsk0;
			end;
		 sum1:=0;
		 sum2:=0;
		 for i:=1 to m_proc do
			begin
			 if i<=n_task then sum1:=sum1+tmp_task[i]^.tasktime;
			 sum2:=sum2+comp_pr[1+(i-1)*k_arr[l]]^.speed;
			 sum:=sum1/sum2;
			 if c<sum then c:=sum;
			end;
		 for i:=m_proc+1 to n_task do sum1:=sum1+tmp_task[i]^.tasktime;
		 sum:=sum1/sum2;
		 if c<sum then c:=sum;
		 for i:=1 to n_task do dispose(tmp_task[i]);
     {release(heapPTR);}
	  end;
  end;
 calculate_C:=c;
end;

procedure compr(var duo:ig;var comp_duo:comp_procs; kp:ig);
var i:ig;
begin
 duo:=duo div kp;
 for i:=1 to duo do comp_duo[i]^:=comp_duo[kp*(i-1)+1]^;
end;


procedure put_c_res(var c:comp_el;nr_t:ig;var sched:schedul_rst;kp:ig);
var i:ig;
begin
 with sched.schedul[c.pr_nr]^ do
  begin
   nrts:=nrts+1;
   with a[nrts] do
    begin
     tnr:=nr_t;
     begint:=c.beg_time;
     endt:=c.end_time;
     usestime:=c.end_time-c.beg_time;
    end;
  end;
 if tasks.task_v[nr_t]^.prcsnum>1 then
 for i:=1 to kp-1 do sched.schedul[c.pr_nr+i]^:=sched.schedul[c.pr_nr]^;
end;


procedure rule1_2_3(var nr_d:ig;var comp_p:comp_procs;
					 nr_w:ig;var w_task:task_arr;
					 var sched:schedul_rst;kp:ig);
var
	i,j,k,l:ig;
	find,findr1:boolean;
	u:real;
	c:comp_el;
	tt:real;


procedure compute_u(var prk,prk1:comp_proc;time:real;nr_t:ig);
var i:ig;nr_c1,nr_c:ig;kdw,k1dw:ig;
    tpc,tpc1:real;speed,speed1:ig;
    beg_time,end_time,u:real;

function tail(var comp_pk:comp_proc;t:real;var nr_c:ig):real;
var i:ig;tpc:real;
begin
 nr_c:=1;
 with comp_pk do
  begin
   tpc:=0;
   while (nr_c<nr_comp) and (comp_a[nr_c].end_time<t) do
    begin
     tpc:=tpc+comp_a[nr_c].comp_pc;
     nr_c:=nr_c+1;
    end;
   with comp_a[nr_c] do if t>beg_time then tail:=pc-tpc-(t-beg_time)*comp_speed
   else tail:=pc;
  end;
end;

function head(var comp_pk:comp_proc;t:real;var nr_c:ig):real;
var i:ig;tpc:real;
begin
 nr_c:=1;
 with comp_pk do
  begin
   tpc:=0;
   while (nr_c<nr_comp) and (comp_a[nr_c].end_time<t) do
    begin
     tpc:=tpc+comp_a[nr_c].comp_pc;
     nr_c:=nr_c+1;
    end;
   with comp_a[nr_c] do head:=tpc+(t-beg_time)*comp_speed;
  end;
end;

procedure connect(var prk,prk1:comp_proc;n,n1:ig;u:real;nr_t:ig);
var i:ig;c:comp_el;
begin
 for i:=1 to n-1 do put_c_res(prk.comp_a[i],nr_t,sched,kp);
 c:=prk.comp_a[n];
 with c do
  begin
   end_time:=u;
   comp_pc:=(end_time-beg_time)*comp_speed;
  end;
 put_c_res(c,nr_t,sched,kp);
 if n1>0 then
  begin
   c:=prk1.comp_a[n1];
   with c do
    begin
     beg_time:=u;
     comp_pc:=(end_time-beg_time)*comp_speed;
    end;
	 put_c_res(c,nr_t,sched,kp);
  end;
 for i:=n1+1 to prk1.nr_comp do put_c_res(prk1.comp_a[i],nr_t,sched,kp);
 with prk1 do
  begin
   if n1>0 then with comp_a[n1] do
    begin
     end_time:=u;
     comp_pc:=(end_time-beg_time)*comp_speed;
    end;
   nr_comp:=n1;
   with prk.comp_a[n] do
    begin
     beg_time:=u;
     comp_pc:=(end_time-beg_time)*comp_speed;
     if comp_pc<re then n:=n+1;
    end;
   for i:=1 to (prk.nr_comp-n+1) do comp_a[nr_comp+i]:=prk.comp_a[n+i-1];
   nr_comp:=nr_comp+(prk.nr_comp-n+1);
   pc:=0;
   for i:=1 to nr_comp do pc:=pc+comp_a[i].comp_pc;
  end;
 prk:=prk1;
end;

begin
 with prk do
  begin
   tpc:=comp_a[1].comp_pc;
   kdw:=1;nr_c:=1;
   while (kdw<nr_comp) and ((tpc+tail(prk1,comp_a[kdw+1].beg_time,nr_c))<time) do
    begin
     kdw:=kdw+1;
     tpc:=tpc+comp_a[kdw].comp_pc;
    end;
   tpc:=tpc-comp_a[kdw].comp_pc;
  end;
  with prk1 do
   begin
    tpc1:=0;
    k1dw:=nr_comp;
    while (k1dw>0) and ((tpc1+comp_a[k1dw].comp_pc+head(prk,comp_a[k1dw].beg_time,nr_c1))>=time) do
     begin
      tpc1:=tpc1+comp_a[k1dw].comp_pc;
      k1dw:=k1dw-1;
     end;
   end;
  if k1dw>0 then
   begin
    speed1:=prk1.comp_a[k1dw].comp_speed;
    end_time:=prk1.comp_a[k1dw].end_time;
   end
  else
   begin
    speed1:=0;
    end_time:=0;
   end;
  beg_time:=prk.comp_a[kdw].beg_time;
  speed:=prk.comp_a[kdw].comp_speed;
  {if speed=speed1 then anykey;}
  u:=(time-tpc-tpc1+speed*beg_time-speed1*end_time)/(speed-speed1);
  connect(prk,prk1,kdw,k1dw,u,nr_t);
end;  { compute_u }

begin
 while nr_w>0 do
  begin
   k:=0;
   findr1:=false;
   while (k<nr_d) and not findr1 do
    begin
     k:=k+1;
     if (abs(comp_p[k]^.pc-w_task[1]^.tasktime)<re) and
     ((k=nr_d) or (abs(comp_p[k+1]^.pc-w_task[1]^.tasktime)>0.001)) then
     findr1:=true;
    end;
   if findr1 then            { rule 1 }
		begin
		{if k=nr_list then join_list(k-1) else}
		 join_list(k);
     with comp_p[k]^ do
      begin
       for j:=1 to nr_comp-1 do
       for l:=j+1 to nr_comp do
       if comp_a[j].beg_time>comp_a[l].beg_time then
        begin
         c:=comp_a[j];comp_a[j]:=comp_a[l];comp_a[l]:=c
        end;
       while (nr_comp>0) do
        begin
				 put_c_res(comp_a[1],w_task[1]^.nr,sched,kp);
         for l:=1 to nr_comp-1 do
         comp_a[l]:=comp_a[l+1];
         nr_comp:=nr_comp-1;
        end;
       for j:=1 to nr_w-1 do w_task[j]^:=w_task[j+1]^;
       nr_w:=nr_w-1;
       for j:=k to nr_d-1 do comp_p[j]^:=comp_p[j+1]^;
       nr_d:=nr_d-1;
      end;
    end
   else
    begin
     k:=0;
     find:=false;
     while (k<nr_d-1) and not find  do
      begin
       k:=k+1;
       if (comp_p[k]^.pc>w_task[1]^.tasktime) and (comp_p[k+1]^.pc<w_task[1]^.tasktime) then find:=true;
      end;
     if find then            { rule 2 }
		  begin
			 join_list(k);
       compute_u(comp_p[k]^,comp_p[k+1]^,w_task[1]^.tasktime,w_task[1]^.nr);
       for j:=1 to nr_w-1 do w_task[j]^:=w_task[j+1]^;
       nr_w:=nr_w-1;
       for j:=k to nr_d-1 do comp_p[j]^:=comp_p[j+1]^;
       nr_d:=nr_d-1;
      end
     else                    { rule 3 }
     with comp_p[nr_d]^ do
			begin
			 if nr_d=1 then join_list(1);
			 for j:=1 to nr_comp-1 do
       for k:=j+1 to nr_comp do
       if comp_a[j].beg_time>comp_a[k].beg_time then
        begin
         c:=comp_a[j];comp_a[j]:=comp_a[k];comp_a[k]:=c
        end;
       tt:=w_task[1]^.tasktime;
       while (nr_comp>0) and (tt>0) do if tt>=comp_a[1].comp_pc then
				begin
				 put_c_res(comp_a[1],w_task[1]^.nr,sched,kp);
         pc:=pc-comp_a[1].comp_pc;
         tt:=tt-comp_a[1].comp_pc;
         for k:=1 to nr_comp-1 do comp_a[k]:=comp_a[k+1];
         nr_comp:=nr_comp-1;
        end
       else
        begin
         c:=comp_a[1];
         c.end_time:=c.beg_time+tt/c.comp_speed;
         c.comp_pc:=tt;
				 put_c_res(c,w_task[1]^.nr,sched,kp);
         comp_a[1].beg_time:=c.end_time;
         comp_a[1].comp_pc:=comp_a[1].comp_pc-tt;
         pc:=pc-tt;
         tt:=0;
        end;
       for j:=1 to nr_w-1 do w_task[j]^:=w_task[j+1]^;
       nr_w:=nr_w-1;
       if nr_comp=0 then nr_d:=nr_d-1;
      end;
    end;
  end;
end; { rule 123 }


procedure decompr(var duo:ig;var comp_duo:comp_procs; kp:ig);
var i,j,k:ig;
    t_c:comp_procs;
    heapret:pointer;
begin
 for i:=1 to comp.nr_comp_proc do
  begin
   new(t_c[i]);
   Inc(memory,SizeOf(comp_proc)+SizeOf(t_c[i]));
   t_c[i]^:=comp_duo[i]^;
  end;
 for i:=1 to duo do
 for j:=1 to kp do
  begin
   comp_duo[kp*(i-1)+j]^:=t_c[i]^;
   with comp_duo[kp*(i-1)+j]^ do
    begin
     proc_nr:=kp*(i-1)+j;
		 for k:=1 to nr_comp do comp_a[k].pr_nr:=comp_a[k].pr_nr+j-1;
    end;
  end;
 for i:=1 to comp.nr_comp_proc do dispose(t_c[i]);
 duo:=duo*kp;
end;

procedure copy_comp(c_from:comp_rec; var c_to:comp_procs);
var
	i:ig;
begin
	for i:=1 to c_from.nr_comp_proc do
		c_to[i]^:=c_from.comp_pr[i]^;
end;

procedure copy_task(tasks:task_rec; var k_task:task_arr; k:ig);
var
	i:ig;
begin
 for i:=1 to nr_t_gr[k] do
  k_task[i]^:=tasks.task_v[tsk(tasks,i,k_arr[k])]^;
end;


procedure dp(var nr_d:ig;var comp_p:comp_procs;
             var nr_t:ig;var t_task:task_arr;
             var sched:schedul_rst;
    	     var x,nrpd:ig;var tab_nrdp:nrdp_arr;var tab_dp:dp_arr;k:ig);
var g,i,j:ig;
    tdp,tt,tpc:real;tx:ig;

function r_sch(x:ig):real;
var i:ig;
    t:real;
begin
 t:=0;
 if x>0 then
  begin
   with sched do
   with schedul[x]^ do
   for i:=1 to nrts do t:=t+a[i].usestime;
  end;
 r_sch:=sched.c-t;
end;

begin
 tdp:=-999999;
 tt:=0;tpc:=0;x:=0;nrpd:=0;
 g:=0;
 while (g<nr_d) and (g<nr_t) do
  begin
   g:=g+1;
   tt:=tt+t_task[g]^.tasktime;
   tpc:=tpc+comp_p[g]^.pc;
   if (tt-tpc)>re then
    begin
     tdp:=tt-tpc;
     if tdp>re then
      begin
       x:=comp_p[g]^.comp_a[comp_p[g]^.nr_comp].pr_nr;
       nrpd:=nrpd+1;
       tab_dp[nrpd]:=tdp;
       tab_nrdp[nrpd]:=comp_p[g]^.comp_a[1].pr_nr;
       tt:=0;tpc:=0;
      end;
    end;
  end;
 if x=0 then
  begin
   j:=g+1;
   while (j<=nr_t) do
    begin
     tt:=tt+t_task[j]^.tasktime;
     if (tt-tpc)>re then
      begin
       tdp:=tt-tpc;
       if tdp>re then
        begin
         x:=comp_p[g]^.comp_a[comp_p[g]^.nr_comp].pr_nr;
         nrpd:=nrpd+1;
         tab_dp[nrpd]:=tdp;
         tab_dp[nrpd]:=comp_p[g]^.comp_a[1].pr_nr;
         tt:=0;tpc:=0;
        end;
      end;
     j:=j+1;
    end;
  end;
 if x>0 then
  begin
	 x:=(x div k_arr[k+1])*k_arr[k+1]+1;
   while (x>1) and (abs(r_sch(x-1))<re) do x:=x-1;
  end;
end; { dp }


function delta(x,nrpd:ig;var tab_nrdp:nrdp_arr;var tab_dp:dp_arr;
					var c:comp_rec;duo:ig;var comp_p:comp_procs;k:ig):real;
var
 i,endp,g,j,q,y,t:ig;
 tsu:real;
 tsd:lig;
 fin:boolean;
 max_delta,temp_delta:real;
 wsk:head_list_ptr;

begin
 max_delta:=0;
 tsd:=0;
 with c do
  begin
	 for g:=1 to nrpd do
		begin
		 tsu:=0;
		 for i:=1 to g do tsu:=tsu+tab_dp[i];
		 t:=tab_nrdp[g];
		 for i:=1 to duo do
		 with comp_p[i]^ do
		 for j:=1 to nr_comp do if comp_a[j].pr_nr=t then y:=i;
		 wsk:=proc_arr[y];
		 while wsk<>nil do
			begin
			 tsd:=tsd+comp_pr[wsk^.first_proc]^.speed;
			 wsk:=wsk^.next_h_l;
			end;
		 temp_delta:=tsu/tsd;
		 if temp_delta>max_delta then max_delta:=temp_delta;
		end;
	end;
 delta:=max_delta;
end;                  { delta }


begin
 memory:=0;
 for i:=1 to MaxM do
  begin
   proc_arr[i]:=nil;
   tmp_comp[i]:=nil;
  end;
 for i:=1 to MaxN do k_task[i]:=nil;
 gettime(hr,mn,sc,sc100);
 with comp do						{ Sorting processors by speed}
  begin
   for i:=1 to nr_comp_proc-1 do
   for j:=i+1 to nr_comp_proc do
    if comp_pr[j]^.speed>comp_pr[i]^.speed then
     begin
      tmp_proc:=comp_pr[i]^;
      comp_pr[i]^:=comp_pr[j]^;
      comp_pr[j]^:=tmp_proc;
     end;
    for i:=1 to nr_comp_proc do
     with comp_pr[i]^ do
      begin
       proc_nr:=i;
       comp_a[1].pr_nr:=i;
      end;
  end;
  with tasks do				{ Sorting tasks by execution time}
   begin
    for i:=1 to n-1 do
    for j:=i+1 to n do
      if task_v[j]^.tasktime>task_v[i]^.tasktime then
	begin
	 tmp_task:=task_v[i]^;
	 task_v[i]^:=task_v[j]^;
	 task_v[j]^:=tmp_task;
	end;
      for i:=1 to n do task_v[i]^.nr:=i;
  end;
 sched.m:=comp.nr_comp_proc;
 for i:=1 to tasks.n do
  begin
   new(k_task[i]);
   Inc(memory,SizeOf(task_rec0)+SizeOf(k_task[i]));
  end;
 for i:=1 to comp.nr_comp_proc do
  begin
   new(tmp_comp[i]);
   Inc(memory,SizeOf(comp_proc)+SizeOf(tmp_comp[i]));
   new(sched.schedul[i]);
  end;
 setrst(sched,comp);
 nr_list:=0;
 init_list(comp.nr_comp_proc);
 nr_t_gr[0]:=0;
 chack_tasks(tasks);
 cmax:=calculate_C(tasks,comp);
 sched.c:=cmax;
 copy_comp(comp,tmp_comp);
 m_comp:=comp.nr_comp_proc;
 cap_end_time(cmax,m_comp,tmp_comp);
 compr(m_comp,tmp_comp,k_arr[nr_k_arr]);
 copy_task(tasks,k_task,nr_k_arr);
 split_list(k_arr[nr_k_arr]);
 rule1_2_3(m_comp,tmp_comp,nr_t_gr[nr_k_arr],k_task,sched,k_arr[nr_k_arr]);
 decompr(m_comp,tmp_comp,k_arr[nr_k_arr]);
 nr_cmax:=1;
 a_cmax[nr_cmax]:=cmax;
 for tmp_k:=nr_k_arr-1 downto 1 do
  begin
   a_cmax[nr_cmax]:=cmax;
   copy_task(tasks,k_task,tmp_k);
   dp(m_comp,tmp_comp,nr_t_gr[tmp_k],k_task,sched,x,nrpd,tab_nrdp,tab_dp,tmp_k);
   if x>0 then
    while x>0 do
     begin
      inc(nr_cmax);
      split_list(k_arr[tmp_k]);
      cmax:=cmax+delta(x,nrpd,tab_nrdp,tab_dp,comp,m_comp,tmp_comp,tmp_k);
      a_cmax[nr_cmax]:=cmax;
      copy_comp(comp,tmp_comp);
      m_comp:=comp.nr_comp_proc;
      cap_end_time(cmax,m_comp,tmp_comp);
      setrst(sched,comp);
      init_list(comp.nr_comp_proc);
      for j:=nr_k_arr downto tmp_k do
	 begin
          split_list(k_arr[j]);
	  sched.c:=cmax;
	  copy_task(tasks,k_task,j);
	  compr(m_comp,tmp_comp,k_arr[j]);
	  rule1_2_3(m_comp,tmp_comp,nr_t_gr[j],k_task,sched,k_arr[j]);
	  decompr(m_comp,tmp_comp,k_arr[j]);
        end;
      if tmp_k>1 then copy_task(tasks,k_task,tmp_k-1)
      else copy_task(tasks,k_task,1);
	 dp(m_comp,tmp_comp,nr_t_gr[j-1],k_task,sched,x,nrpd,tab_nrdp,tab_dp,tmp_k);
     end
   else
    begin
     compr(m_comp,tmp_comp,k_arr[tmp_k]);
     split_list(k_arr[tmp_k]);
     rule1_2_3(m_comp,tmp_comp,nr_t_gr[tmp_k],k_task,sched,k_arr[tmp_k]);
     decompr(m_comp,tmp_comp,k_arr[tmp_k]);
    end;
  end;
 workt:=getsec(hr,mn,sc,sc100);

	{ The end of scheduling algorithm }
 with sched do
 for i:=1 to m do
 with schedul[i]^ do
 for j:=1 to nrts-1 do
 for l:=j+1 to nrts do if a[j].begint>a[l].begint then
  begin
   t_ra:=a[j];
	 a[j]:=a[l];
	 a[l]:=t_ra;
  end;
 for i:=1 to MaxM do
  begin
   if tmp_comp[i]<>nil then dispose(tmp_comp[i]);
   while proc_arr[i]<>nil do
    begin
     stake:=proc_arr[i];
     proc_arr[i]:=proc_arr[i]^.next_h_l;
     dispose(stake);
    end;
  end;
 for i:=1 to MaxN do if k_task[i]<>nil then dispose(k_task[i]);
end;

end.
